#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2019-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
from typing import Any, Optional

import numpy as np

import rapidtide.io as tide_io
import rapidtide.multiproc as tide_multiproc
import rapidtide.workflows.parser_funcs as pf

try:
    import tensorflow as tf

    tensorflowpresent = True
except ImportError:
    tensorflowpresent = False

DEFAULT_ALIASEDCORRELATIONWIDTH = 5.0
DEFAULT_PULSATILITYSIGMA = 6.0
DEFAULT_PULSATILITYTHRESHOLD = 0.5
DEFAULT_TF_DL_MODEL = "model_revised_tf2"
DEFAULT_PT_DL_MODEL = "model_ppgattention_pytorch_w128_fulldata"


def _get_parser() -> Any:
    """
    Create and configure an argument parser for the happy command-line tool.

    This function sets up an `argparse.ArgumentParser` with numerous options for
    configuring the hypersampling by analytic phase projection (HAPPY) pipeline.
    It includes arguments for input files, processing steps, performance tuning,
    preprocessing, cardiac and respiration estimation, output handling, and debugging.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser with all supported command-line options.

    Notes
    -----
    The parser includes multiple argument groups for different functional areas:
    - Processing steps
    - Performance
    - Preprocessing
    - Cardiac estimation tuning
    - External cardiac waveform options
    - External respiration waveform options
    - Output processing
    - Output options
    - Phase projection tuning
    - Miscellaneous options
    - Debugging options

    Examples
    --------
    >>> parser = _get_parser()
    >>> args = parser.parse_args()
    """
    parser = argparse.ArgumentParser(
        prog="happy",
        description="Hypersampling by Analytic Phase Projection - Yay!.",
        allow_abbrev=False,
    )

    # Required arguments
    parser.add_argument(
        "fmrifilename",
        type=lambda x: pf.is_valid_file(parser, x),
        help="The input data file (BOLD fmri file or NIRS text file)",
    )
    parser.add_argument(
        "slicetimename",
        type=lambda x: pf.is_valid_file(parser, x),
        help=(
            "Text file containing the offset time in seconds of each slice relative "
            "to the start of the TR, one value per line, OR the BIDS sidecar JSON file."
            "NB: FSL slicetime files give slice times in fractions of a TR, BIDS sidecars "
            "give slice times in seconds. Non-json files are assumed to be the FSL style "
            "(fractions of a TR) UNLESS the --slicetimesareinseconds flag is used."
        ),
    )
    parser.add_argument("outputroot", help="The root name for the output files")

    # Processing steps
    processing_steps = parser.add_argument_group("Processing steps")
    processing_steps.add_argument(
        "--cardcalconly",
        dest="cardcalconly",
        action="store_true",
        help="Stop after all cardiac regressor calculation steps (before phase projection). ",
        default=False,
    )
    processing_steps.add_argument(
        "--skipdlfilter",
        dest="dodlfilter",
        action="store_false",
        help="Disable deep learning cardiac waveform filter.  ",
        default=True,
    )
    if tensorflowpresent:
        processing_steps.add_argument(
            "--usesuperdangerousworkaround",
            dest="mpfix",
            action="store_true",
            help=(
                "Some versions of tensorflow seem to have some weird conflict with MKL which"
                "I don't seem to be able to fix.  If the dl filter bombs complaining about "
                "multiple openmp libraries, try rerunning with the secret and inadvisable "
                "'--usesuperdangerousworkaround' flag.  Good luck! "
            ),
            default=False,
        )
    processing_steps.add_argument(
        "--slicetimesareinseconds",
        action="store_true",
        help=(
            "If a non-json slicetime file is specified, happy assumes the file is FSL style "
            "(slice times are specified in fractions of a TR).  Setting this flag overrides this "
            "assumption, and interprets the slice time file as being in seconds.  This does "
            "nothing when the slicetime file is a .json BIDS sidecar."
        ),
        default=False,
    )
    processing_steps.add_argument(
        "--model",
        dest="modelname",
        metavar="MODELNAME",
        help=(
            f"Use model MODELNAME for dl filter (default is {DEFAULT_PT_DL_MODEL} - "
            "from the revised NeuroImage paper.) "
        ),
        default=None,
    )

    # Performance
    performance_opts = parser.add_argument_group("Performance")
    performance_opts.add_argument(
        "--mklthreads",
        dest="mklthreads",
        action="store",
        metavar="NTHREADS",
        type=lambda x: pf.is_int(parser, x),
        help=(
            "Use NTHREADS MKL threads to accelerate processing (defaults to 1 - more "
            "threads up to the number of cores can accelerate processing a lot, but "
            "can really kill you on clusters unless you're very careful.  Use at your own risk"
        ),
        default=1,
    )
    performance_opts.add_argument(
        "--nprocs",
        dest="nprocs",
        action="store",
        metavar="NPROCS",
        type=lambda x: pf.is_int(parser, x),
        help=(
            "Use NPROCS CPUs to accelerate processing (defaults to 1 - more "
            "CPUs up to the number of cores can accelerate processing a lot, but "
            "you need to remember to ask for this many CPUs on clusters.)  Entering a mulitprocessor "
            "routine disables mklthreads (otherwise there's chaos)."
        ),
        default=1,
    )
    # Preprocessing
    preprocessing_opts = parser.add_argument_group("Preprocessing")
    preprocessing_opts.add_argument(
        "--numskip",
        dest="numskip",
        action="store",
        metavar="SKIP",
        type=lambda x: pf.is_int(parser, x),
        help="Skip SKIP tr's at the beginning of the fMRI file (default is 0). ",
        default=0,
    )
    preprocessing_opts.add_argument(
        "--motskip",
        dest="motskip",
        action="store",
        metavar="SKIP",
        type=lambda x: pf.is_int(parser, x),
        help="Skip SKIP tr's at the beginning of the motion regressor file (default is 0). ",
        default=0,
    )
    preprocessing_opts.add_argument(
        "--motionfile",
        dest="motionfilespec",
        metavar="MOTFILE",
        help=(
            "Read 6 columns of motion regressors out of MOTFILE file (.par or BIDS .json) "
            "(with timepoints rows) and regress them, their derivatives, "
            "and delayed derivatives out of the data prior to analysis. "
        ),
        default=None,
    )
    preprocessing_opts.add_argument(
        "--motionhp",
        dest="motionhp",
        action="store",
        metavar="HPFREQ",
        type=lambda x: pf.is_float(parser, x),
        help="Highpass filter motion regressors to HPFREQ Hz prior to regression. ",
        default=None,
    )
    preprocessing_opts.add_argument(
        "--motionlp",
        dest="motionlp",
        action="store",
        metavar="LPFREQ",
        type=lambda x: pf.is_float(parser, x),
        help="Lowpass filter motion regressors to LPFREQ Hz prior to regression. ",
        default=None,
    )
    preprocessing_opts.add_argument(
        "--nomotorthogonalize",
        dest="orthogonalize",
        action="store_false",
        help=(
            "Do not orthogonalize motion regressors prior to regressing them out of the " "data. "
        ),
        default=True,
    )
    preprocessing_opts.add_argument(
        "--nomotderiv",
        dest="motfilt_deriv",
        action="store_false",
        help=("Do not use motion derivative regressors. "),
        default=True,
    )
    preprocessing_opts.add_argument(
        "--motfiltorder",
        dest="motfilt_order",
        action="store",
        metavar="N",
        type=lambda x: pf.is_int(parser, x),
        help=(
            "Include powers of each confound regressor up to order N. Default is 1 (no expansion). "
        ),
        default=1,
    )
    preprocessing_opts.add_argument(
        "--discardmotionfiltered",
        dest="savemotionglmfilt",
        action="store_false",
        help=("Do not save data after motion filtering. "),
        default=True,
    )

    # Cardiac estimation tuning
    cardiac_est_tuning = parser.add_argument_group("Cardiac estimation tuning")
    cardiac_est_tuning.add_argument(
        "--estweights",
        dest="estweightsname",
        action="store",
        metavar="WEIGHTSNAME",
        help=(
            "Generation of cardiac waveform from data will be restricted to "
            "voxels in MASKNAME and weighted by the mask intensity.  If this is "
            "selected, happy will only make a single pass through the data (the "
            "initial vessel mask generation pass will be skipped)."
        ),
        default=None,
    )
    cardiac_est_tuning.add_argument(
        "--minhr",
        dest="minhr",
        action="store",
        metavar="MINHR",
        type=lambda x: pf.is_float(parser, x),
        help="Limit lower cardiac frequency search range to MINHR BPM (default is 40). ",
        default=40.0,
    )
    cardiac_est_tuning.add_argument(
        "--maxhr",
        dest="maxhr",
        action="store",
        metavar="MAXHR",
        type=lambda x: pf.is_float(parser, x),
        help="Limit upper cardiac frequency search range to MAXHR BPM (default is 140). ",
        default=140.0,
    )
    cardiac_est_tuning.add_argument(
        "--minhrfilt",
        dest="minhrfilt",
        action="store",
        metavar="MINHR",
        type=lambda x: pf.is_float(parser, x),
        help="Highpass filter cardiac waveform estimate to MINHR BPM (default is 40). ",
        default=40.0,
    )
    cardiac_est_tuning.add_argument(
        "--maxhrfilt",
        dest="maxhrfilt",
        action="store",
        metavar="MAXHR",
        type=lambda x: pf.is_float(parser, x),
        help="Lowpass filter cardiac waveform estimate to MAXHR BPM (default is 1000). ",
        default=1000.0,
    )
    cardiac_est_tuning.add_argument(
        "--hilbertcomponents",
        dest="hilbertcomponents",
        action="store",
        metavar="NCOMPS",
        type=lambda x: pf.is_int(parser, x),
        help="Retain NCOMPS components of the cardiac frequency signal to Hilbert transform (default is 1). ",
        default=1,
    )
    cardiac_est_tuning.add_argument(
        "--envcutoff",
        dest="envcutoff",
        action="store",
        metavar="CUTOFF",
        type=lambda x: pf.is_float(parser, x),
        help="Lowpass filter cardiac normalization envelope to CUTOFF Hz (default is 0.4 Hz). ",
        default=0.4,
    )
    cardiac_est_tuning.add_argument(
        "--notchwidth",
        dest="notchpct",
        action="store",
        metavar="WIDTH",
        type=lambda x: pf.is_float(parser, x),
        help="Set the width of the notch filter, in percent of the notch frequency (default is 1.5). ",
        default=1.5,
    )
    cardiac_est_tuning.add_argument(
        "--invertphysiosign",
        dest="invertphysiosign",
        action="store_true",
        help=(
            "Invert the waveform extracted from the physiological signal.  "
            "Use this if there is a contrast agent in the blood. "
        ),
        default=False,
    )
    cardiac_est_tuning.add_argument(
        "--teoffset",
        dest="teoffset",
        action="store",
        metavar="TE",
        type=lambda x: pf.is_float(parser, x),
        help="Specify the echo time in seconds.  This is used when combining multiecho data.  Default is 0. ",
        default=None,
    )

    # External cardiac waveform options
    external_cardiac_opts = parser.add_argument_group("External cardiac waveform options")
    external_cardiac_opts.add_argument(
        "--cardiacfile",
        dest="cardiacfilename",
        metavar="FILE[:COL]",
        help=(
            "Read the cardiac waveform from file FILE.  If COL is an integer, "
            "and FILE is a text file, use the COL'th column.  If FILE is a BIDS "
            "format json file, use column named COL. If no file is specified, "
            "estimate the cardiac signal from the fMRI data."
        ),
        default=None,
    )
    cardiac_freq = external_cardiac_opts.add_mutually_exclusive_group()
    cardiac_freq.add_argument(
        "--cardiacfreq",
        dest="inputfreq",
        action="store",
        metavar="FREQ",
        type=lambda x: pf.is_float(parser, x),
        help=(
            "Cardiac waveform in cardiacfile has sample frequency FREQ "
            "(default is 32Hz). NB: --cardiacfreq and --cardiactstep "
            "are two ways to specify the same thing. "
        ),
        default=-32.0,
    )
    cardiac_freq.add_argument(
        "--cardiactstep",
        dest="inputfreq",
        action="store",
        metavar="TSTEP",
        type=lambda x: pf.invert_float(parser, x),
        help=(
            "Cardiac waveform in cardiacfile has time step TSTEP "
            "(default is 1/32 sec). NB: --cardiacfreq and --cardiactstep "
            "are two ways to specify the same thing. "
        ),
        default=-32.0,
    )
    external_cardiac_opts.add_argument(
        "--cardiacstart",
        dest="inputstart",
        metavar="START",
        action="store",
        type=float,
        help=(
            "The time delay in seconds into the cardiac file, corresponding "
            "to the first TR of the fMRI file (default is 0.0) "
        ),
        default=None,
    )
    external_cardiac_opts.add_argument(
        "--forcehr",
        dest="forcedhr",
        metavar="BPM",
        action="store",
        type=lambda x: pf.is_float(parser, x) / 60.0,
        help=(
            "Force heart rate fundamental detector to be centered at BPM "
            "(overrides peak frequencies found from spectrum).  Useful"
            "if there is structured noise that confuses the peak finder. "
        ),
        default=None,
    )

    respiration = True
    if respiration:
        # External respiration waveform options
        external_respiration_opts = parser.add_argument_group(
            "External respiration waveform options"
        )
        external_respiration_opts.add_argument(
            "--respirationfile",
            dest="respirationfilename",
            metavar="FILE[:COL]",
            help=(
                "Read the respiration waveform from file FILE.  If COL is an integer, "
                "and FILE is a text file, use the COL'th column.  If FILE is a BIDS "
                "format json file, use column named COL."
            ),
            default=None,
        )
        respiration_freq = external_respiration_opts.add_mutually_exclusive_group()
        respiration_freq.add_argument(
            "--respirationfreq",
            dest="respinputfreq",
            action="store",
            metavar="FREQ",
            type=lambda x: pf.is_float(parser, x),
            help=(
                "Respiration waveform in respirationfile has sample frequency FREQ "
                "(default is 32Hz). NB: --respirationfreq and --respirationtstep "
                "are two ways to specify the same thing. "
            ),
            default=-32.0,
        )
        respiration_freq.add_argument(
            "--respirationtstep",
            dest="respinputfreq",
            action="store",
            metavar="TSTEP",
            type=lambda x: pf.invert_float(parser, x),
            help=(
                "Respiration waveform in respirationfile has time step TSTEP "
                "(default is 1/32 sec). NB: --respirationfreq and --respirationtstep "
                "are two ways to specify the same thing. "
            ),
            default=-32.0,
        )
        external_respiration_opts.add_argument(
            "--respirationstart",
            dest="respinputstart",
            metavar="START",
            action="store",
            type=float,
            help=(
                "The time delay in seconds into the respiration file, corresponding "
                "to the first TR of the fMRI file (default is 0.0) "
            ),
            default=None,
        )
        external_respiration_opts.add_argument(
            "--forcerr",
            dest="forcedrr",
            metavar="BreathsPM",
            action="store",
            type=lambda x: pf.is_float(parser, x) / 60.0,
            help=(
                "Force respiratory rate fundamental detector to be centered at BreathsPM "
                "(overrides peak frequencies found from spectrum).  Useful"
                "if there is structured noise that confuses the peak finder. "
            ),
            default=None,
        )

    # Output processing
    output_proc = parser.add_argument_group(
        title="Output processing - mutually exclusive options",
        description=(
            "Optional cardiac noise removal.  You can do either spatial or temporal regression, but not both."
        ),
    ).add_mutually_exclusive_group()
    output_proc.add_argument(
        "--spatialregression",
        dest="dospatialregression",
        action="store_true",
        help="Generate framewise cardiac signal maps and filter them out of the input data. ",
        default=False,
    )
    output_proc.add_argument(
        "--temporalregression",
        dest="dotemporalregression",
        action="store_true",
        help="Generate voxelwise aliased synthetic cardiac regressors and filter them out of the input data. ",
        default=False,
    )

    # Output options
    output = parser.add_argument_group("Output options")
    output.add_argument(
        "--stdfreq",
        dest="stdfreq",
        metavar="FREQ",
        action="store",
        type=float,
        help=(
            "Frequency to which the physiological signals are resampled for output. "
            "Default is 25. "
        ),
        default=25.0,
    )

    # Phase projection tuning
    phase_proj_tuning = parser.add_argument_group("Phase projection tuning")
    phase_proj_tuning.add_argument(
        "--outputbins",
        dest="destpoints",
        metavar="BINS",
        action="store",
        type=lambda x: pf.is_int(parser, x),
        help="Number of output phase bins (default is 32). ",
        default=32,
    )
    phase_proj_tuning.add_argument(
        "--gridbins",
        dest="congridbins",
        metavar="BINS",
        action="store",
        type=lambda x: pf.is_float(parser, x),
        help="Width of the gridding kernel in output phase bins (default is 3.0). ",
        default=3.0,
    )
    phase_proj_tuning.add_argument(
        "--gridkernel",
        dest="gridkernel",
        action="store",
        type=str,
        choices=["old", "gauss", "kaiser"],
        help="Convolution gridding kernel. Default is kaiser",
        default="kaiser",
    )
    phase_proj_tuning.add_argument(
        "--projmask",
        dest="projmaskname",
        metavar="MASKNAME",
        help=(
            "Phase projection will be restricted to voxels in MASKNAME "
            "(overrides normal intensity mask.) "
        ),
        default=None,
    )
    phase_proj_tuning.add_argument(
        "--projectwithraw",
        dest="projectwithraw",
        action="store_true",
        help="Use fMRI derived cardiac waveform as phase source for projection, even if a plethysmogram is supplied.",
        default=False,
    )
    phase_proj_tuning.add_argument(
        "--fliparteries",
        dest="fliparteries",
        action="store_true",
        help=(
            "Attempt to detect arterial signals and flip over the timecourses after phase projection "
            "(since relative arterial blood susceptibility is inverted relative to venous blood)."
        ),
        default=False,
    )
    phase_proj_tuning.add_argument(
        "--arteriesonly",
        dest="arteriesonly",
        action="store_true",
        help="Restrict cardiac waveform estimation to putative arteries only.",
        default=False,
    )
    phase_proj_tuning.add_argument(
        "--pulsatilitysigma",
        dest="pulsatilitysigma",
        action="store",
        metavar="SIGMAINMM",
        type=lambda x: pf.is_float(parser, x),
        help=(
            "Split pulsatility map spatial frequencies with a gaussian filter with sigma=SIGMAINMM. "
            f"Default is {DEFAULT_PULSATILITYSIGMA}mm. "
        ),
        default=DEFAULT_PULSATILITYSIGMA,
    )
    phase_proj_tuning.add_argument(
        "--pulsatilitythreshold",
        dest="pulsatilitythreshold",
        action="store",
        metavar="THRESHPCT",
        type=lambda x: pf.is_float(parser, x),
        help=(
            "Consider voxels with a high spatial frequency pulsatility above THRESHPCT percent "
            "to be vessels. "
            f"Default is {DEFAULT_PULSATILITYTHRESHOLD}mm. "
        ),
        default=DEFAULT_PULSATILITYSIGMA,
    )

    # Add version options
    pf.addversionopts(parser)

    # Add miscellaneous options
    misc_opts = parser.add_argument_group("Miscellaneous options.")
    misc_opts.add_argument(
        "--processmask",
        dest="processmask",
        action="store",
        metavar="MASKNAME",
        help=(
            "Instead of dynamically generating a processing mask from the data, use "
            "the mask specified in the file MASKNAME. "
            "The mask must be a 3D NIFTI file with x, y, z dimensions matching the fMRI data. "
        ),
        default=None,
    )
    misc_opts.add_argument(
        "--aliasedcorrelation",
        dest="doaliasedcorrelation",
        action="store_true",
        help="Attempt to calculate absolute delay using an aliased correlation (experimental).",
        default=False,
    )
    misc_opts.add_argument(
        "--aliasedcorrelationwidth",
        dest="aliasedcorrelationwidth",
        metavar="WIDTH",
        action="store",
        type=lambda x: pf.is_float(parser, x),
        help=f"Width of the aliased correlation calculation (default is {DEFAULT_ALIASEDCORRELATIONWIDTH}). ",
        default=DEFAULT_ALIASEDCORRELATIONWIDTH,
    )
    misc_opts.add_argument(
        "--upsample",
        dest="doupsampling",
        action="store_true",
        help="Attempt to temporally upsample the fMRI data (experimental).",
        default=False,
    )
    misc_opts.add_argument(
        "--estimateflow",
        dest="doflowfields",
        action="store_true",
        help="Estimate blood flow using optical flow (experimental).",
        default=False,
    )
    misc_opts.add_argument(
        "--noprogressbar",
        dest="showprogressbar",
        action="store_false",
        help="Will disable showing progress bars (helpful if stdout is going to a file). ",
        default=True,
    )
    misc_opts.add_argument(
        "--wrightiterations",
        dest="wrightiterations",
        action="store",
        type=lambda x: pf.is_int(parser, x),
        help="Number of iterations for calculating Wright map. Set to 0 to disable.",
        default=0,
    )
    misc_opts.add_argument(
        "--useoldvesselmethod",
        dest="useoriginalvesselmethod",
        action="store_true",
        help="Will use the old method to find blood vessels. ",
        default=False,
    )
    pf.addtagopts(
        misc_opts,
        helptext="Additional key, value pairs to add to the info json file (useful for tracking analyses).",
    )

    # Debugging options
    debug_opts = parser.add_argument_group("Debugging options (probably not of interest to users)")
    debug_opts.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="Turn on debugging information.",
        default=False,
    )
    debug_opts.add_argument(
        "--nodetrend",
        dest="detrendorder",
        action="store",
        type=lambda x: pf.is_int(parser, 0),
        help="Disable data detrending. ",
        default=3,
    )
    debug_opts.add_argument(
        "--normvesselmap",
        dest="unnormvesselmap",
        action="store_false",
        help="Find vessels using normalized phase projection. ",
        default=True,
    )
    debug_opts.add_argument(
        "--noorthog",
        dest="orthogonalize",
        action="store_false",
        help="Disable orthogonalization of motion confound regressors. ",
        default=True,
    )
    debug_opts.add_argument(
        "--disablenotch",
        dest="disablenotch",
        action="store_true",
        help="Disable subharmonic notch filter. ",
        default=False,
    )
    debug_opts.add_argument(
        "--nomask",
        dest="usemaskcardfromfmri",
        action="store_false",
        help="Disable data masking for calculating cardiac waveform. ",
        default=True,
    )
    debug_opts.add_argument(
        "--nocensor",
        dest="censorbadpts",
        action="store_false",
        help="Bad points will not be excluded from analytic phase projection. ",
        default=True,
    )
    debug_opts.add_argument(
        "--noappsmooth",
        dest="smoothapp",
        action="store_false",
        help="Disable smoothing app file in the phase direction. ",
        default=True,
    )
    debug_opts.add_argument(
        "--nophasefilt",
        dest="filtphase",
        action="store_false",
        help="Disable the phase trend filter (probably not a good idea). ",
        default=True,
    )
    debug_opts.add_argument(
        "--nocardiacalign",
        dest="aligncardiac",
        action="store_false",
        help="Disable alignment of pleth signal to fMRI derived cardiac signal. ",
        default=True,
    )
    debug_opts.add_argument(
        "--saveinfoastext",
        dest="saveinfoasjson",
        action="store_false",
        help="Save the info file in text format rather than json. ",
        default=True,
    )
    debug_opts.add_argument(
        "--saveintermediate",
        dest="saveintermediate",
        action="store_true",
        help="Save some data from intermediate passes to help debugging. ",
        default=False,
    )
    debug_opts.add_argument(
        "--increaseoutputlevel",
        dest="inc_outputlevel",
        action="count",
        help="Increase the number of intermediate output files. ",
        default=0,
    )
    debug_opts.add_argument(
        "--decreaseoutputlevel",
        dest="dec_outputlevel",
        action="count",
        help="Decrease the number of intermediate output files. ",
        default=0,
    )
    debug_opts.add_argument(
        "--nompdetrend",
        dest="mpdetrend",
        action="store_false",
        help="Disable multiproc detrending.",
        default=True,
    )
    debug_opts.add_argument(
        "--nompphaseproject",
        dest="mpphaseproject",
        action="store_false",
        help="Disable multiproc phase projection.",
        default=True,
    )
    debug_opts.add_argument(
        "--noprefillcongrid",
        dest="preloadcongrid",
        action="store_false",
        help="Don't prefill the congrid value cache.",
        default=True,
    )
    debug_opts.add_argument(
        "--nocongridcache",
        dest="congridcache",
        action="store_false",
        help="Disable the congrid value cache completely.",
        default=True,
    )
    if tensorflowpresent:
        debug_opts.add_argument(
            "--usetensorflow",
            dest="usepytorch",
            action="store_false",
            help=("Switch to the old style tensorflow deep learning filter"),
            default=True,
        )
    debug_opts.add_argument(
        "--focaldebug",
        dest="focaldebug",
        action="store_true",
        help=("Enable targeted additional debugging output (used during development)."),
        default=False,
    )

    return parser


def process_args(inputargs: Optional[Any] = None) -> Any:
    """
    Compile arguments for happy workflow.

    This function processes input arguments for the happy workflow, handling
    argument parsing, command line formatting, default model selection, and post-processing
    of various parameters. It also manages motion file specifications, multiprocessing
    settings, and output level calculations.

    Parameters
    ----------
    inputargs : optional
        Input arguments to be processed. If None, arguments are parsed from the command line.
        Default is None.

    Returns
    -------
    args : argparse.Namespace
        Processed arguments namespace containing all parsed and defaulted values.

    Notes
    -----
    - If no model is specified, the function sets a default based on whether PyTorch or TensorFlow
      is being used.
    - The function writes formatted command line arguments to files for logging purposes.
    - Multiprocessing is currently disabled due to instability.
    - The output level is adjusted based on user-provided increments and decrements.

    Examples
    --------
    >>> args = process_args()
    >>> print(args.outputroot)
    >>> print(args.modelname)
    """
    args, argstowrite = pf.setargs(_get_parser, inputargs=inputargs)

    # save the raw and formatted command lines
    args.commandline = " ".join(argstowrite)
    tide_io.writevec(np.array([args.commandline]), args.outputroot + "_commandline.txt")
    formattedcommandline = []
    for thetoken in argstowrite[0:3]:
        formattedcommandline.append(thetoken)
    for thetoken in argstowrite[3:]:
        if thetoken[0:2] == "--":
            formattedcommandline.append(thetoken)
        else:
            formattedcommandline[-1] += " " + thetoken
    for i in range(len(formattedcommandline)):
        if i > 0:
            prefix = "    "
        else:
            prefix = ""
        if i < len(formattedcommandline) - 1:
            suffix = " \\"
        else:
            suffix = ""
        formattedcommandline[i] = prefix + formattedcommandline[i] + suffix
    tide_io.writevec(np.array(formattedcommandline), args.outputroot + "_formattedcommandline.txt")

    if not tensorflowpresent:
        args.usepytorch = True
        args.mpfix = False

    # if user did not specify a model, set the default, depending on DL library
    if args.modelname is None:
        if args.usepytorch:
            args.modelname = DEFAULT_PT_DL_MODEL
        else:
            args.modelname = DEFAULT_TF_DL_MODEL
        if args.debug:
            print(f"No model specified.  {args.usepytorch=}, so using {args.modelname}")

    if args.debug:
        print()
        print("before postprocessing")
        print(args)

    # some tunable parameters
    args.outputlevel = 1
    args.maskthreshpct = 10.0
    args.domadnorm = True
    args.verbose = False
    args.smoothlen = 101
    args.envthresh = 0.2
    args.upsamplefac = 100
    args.centric = True
    args.pulsereconstepsize = 0.01
    args.histlen = 100
    args.softvesselfrac = 0.4
    args.savecardiacnoise = True
    args.colnum = None
    args.colname = None
    args.notchrolloff = 0.5

    # Additional argument parsing not handled by argparse
    # deal with notch filter logic
    if args.disablenotch:
        args.notchpct = None

    # process motionfile information
    if args.motionfilespec is not None:
        args.motionfilename, args.motionfilecolspec = tide_io.parsefilespec(args.motionfilespec)
    else:
        args.motionfilename = None

    # set the number of worker processes if multiprocessing
    # Multiprocessing is very broken atm.  Hard disable it.
    # args.nprocs = 1
    if args.nprocs < 1:
        args.nprocs = tide_multiproc.maxcpus()

    # process infotags
    args = pf.postprocesstagopts(args)

    # determine the outputlevel
    args.outputlevel = np.max([0, args.outputlevel + args.inc_outputlevel - args.dec_outputlevel])

    if args.debug:
        print()
        print("after postprocessing")
        print(args)

    # start the clock!
    # tide_util.checkimports(args)

    return args



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray

import rapidtide.fit as tide_fit
import rapidtide.io as tide_io
import rapidtide.miscmath as tide_math
import rapidtide.refineregressor as tide_refineregressor
import rapidtide.resample as tide_resample
import rapidtide.stats as tide_stats
import rapidtide.util as tide_util
from rapidtide.tests.utils import mse

"""
A class to implement regressor refinement
"""


class RegressorRefiner:

    refinemaskvoxels = None

    def __init__(
        self,
        internalvalidfmrishape: Any,
        internalvalidpaddedfmrishape: Any,
        pid: Any,
        outputname: Any,
        initial_fmri_x: Any,
        paddedinitial_fmri_x: Any,
        os_fmri_x: Any,
        sharedmem: bool = False,
        offsettime: float = 0.0,
        ampthresh: float = 0.3,
        lagminthresh: float = 0.25,
        lagmaxthresh: float = 3.0,
        sigmathresh: float = 1000.0,
        cleanrefined: bool = False,
        bipolar: bool = False,
        fixdelay: bool = False,
        includemask: Optional[Any] = None,
        excludemask: Optional[Any] = None,
        LGR: Optional[Any] = None,
        nprocs: int = 1,
        detrendorder: int = 1,
        alwaysmultiproc: bool = False,
        showprogressbar: bool = True,
        chunksize: int = 50000,
        padtrs: int = 10,
        refineprenorm: str = "var",
        refineweighting: Optional[Any] = None,
        refinetype: str = "pca",
        pcacomponents: float = 0.8,
        dodispersioncalc: bool = False,
        dispersioncalc_lower: float = -5.0,
        dispersioncalc_upper: float = 5.0,
        dispersioncalc_step: float = 0.5,
        windowfunc: str = "hamming",
        passes: int = 3,
        maxpasses: int = 15,
        convergencethresh: Optional[Any] = None,
        interptype: str = "univariate",
        usetmask: bool = False,
        tmask_y: Optional[Any] = None,
        tmaskos_y: Optional[Any] = None,
        fastresamplerpadtime: float = 45.0,
        prewhitenregressor: bool = False,
        prewhitenlags: int = 10,
        debug: bool = False,
        rt_floattype: np.dtype = np.float64,
    ) -> None:
        """
        Initialize the object with configuration parameters for fMRI data processing.

        This constructor sets up internal attributes and performs initial setup tasks such as
        allocating memory and applying masks based on provided parameters.

        Parameters
        ----------
        internalvalidfmrishape : Any
            Shape of the internal valid fMRI data.
        internalvalidpaddedfmrishape : Any
            Shape of the padded internal valid fMRI data.
        pid : Any
            Process identifier used for memory allocation.
        outputname : Any
            Name of the output file or dataset.
        initial_fmri_x : Any
            Initial fMRI data array.
        paddedinitial_fmri_x : Any
            Padded version of the initial fMRI data.
        os_fmri_x : Any
            Oversampled fMRI data array.
        sharedmem : bool, optional
            Whether to use shared memory for processing (default is False).
        offsettime : float, optional
            Time offset in seconds (default is 0.0).
        ampthresh : float, optional
            Amplitude threshold for filtering (default is 0.3).
        lagminthresh : float, optional
            Minimum lag threshold for correlation analysis (default is 0.25).
        lagmaxthresh : float, optional
            Maximum lag threshold for correlation analysis (default is 3.0).
        sigmathresh : float, optional
            Significance threshold for statistical tests (default is 1000.0).
        cleanrefined : bool, optional
            Whether to apply refined cleaning steps (default is False).
        bipolar : bool, optional
            Whether to use bipolar filtering (default is False).
        fixdelay : bool, optional
            Whether to fix delay in the processing pipeline (default is False).
        includemask : Optional[Any], optional
            Mask to include specific regions (default is None).
        excludemask : Optional[Any], optional
            Mask to exclude specific regions (default is None).
        LGR : Optional[Any], optional
            Logarithmic gradient or related parameter (default is None).
        nprocs : int, optional
            Number of processes to use (default is 1).
        detrendorder : int, optional
            Order of detrending polynomial (default is 1).
        alwaysmultiproc : bool, optional
            Force multiprocessing even for small tasks (default is False).
        showprogressbar : bool, optional
            Show progress bar during processing (default is True).
        chunksize : int, optional
            Size of data chunks for processing (default is 50000).
        padtrs : int, optional
            Number of TRs to pad (default is 10).
        refineprenorm : str, optional
            Pre-normalization method for refinement ("var", "mean", etc.) (default is "var").
        refineweighting : Optional[Any], optional
            Weighting scheme for refinement (default is None).
        refinetype : str, optional
            Type of refinement to perform ("pca", "ica", etc.) (default is "pca").
        pcacomponents : float, optional
            Fraction of PCA components to retain (default is 0.8).
        dodispersioncalc : bool, optional
            Whether to perform dispersion calculation (default is False).
        dispersioncalc_lower : float, optional
            Lower bound for dispersion calculation (default is -5.0).
        dispersioncalc_upper : float, optional
            Upper bound for dispersion calculation (default is 5.0).
        dispersioncalc_step : float, optional
            Step size for dispersion calculation (default is 0.5).
        windowfunc : str, optional
            Window function used in spectral analysis (default is "hamming").
        passes : int, optional
            Number of filter passes (default is 3).
        maxpasses : int, optional
            Maximum allowed number of passes (default is 15).
        convergencethresh : Optional[Any], optional
            Convergence threshold for iterative algorithms (default is None).
        interptype : str, optional
            Interpolation type for resampling ("univariate", "multivariate") (default is "univariate").
        usetmask : bool, optional
            Whether to use temporal mask (default is False).
        tmask_y : Optional[Any], optional
            Temporal mask for y-axis (default is None).
        tmaskos_y : Optional[Any], optional
            Oversampled temporal mask for y-axis (default is None).
        fastresamplerpadtime : float, optional
            Padding time for fast resampling (default is 45.0).
        prewhitenregressor : bool, optional
            Apply pre-whitening to regressors (default is False).
        prewhitenlags : int, optional
            Number of lags for pre-whitening (default is 10).
        debug : bool, optional
            Enable debug mode (default is False).
        rt_floattype : np.dtype, optional
            Rapidtide floating-point data type (default is np.float64).

        Returns
        -------
        None
            This method initializes the object and does not return any value.

        Notes
        -----
        - The function internally calls `setmasks` and `_allocatemem` to initialize
          masks and allocate memory respectively.
        - The parameters are stored as instance attributes for use in subsequent processing steps.

        Examples
        --------
        >>> obj = MyClass(
        ...     internalvalidfmrishape=(64, 64, 30),
        ...     internalvalidpaddedfmrishape=(64, 64, 35),
        ...     pid=12345,
        ...     outputname="output.nii",
        ...     initial_fmri_x=np.random.rand(64, 64, 30),
        ...     paddedinitial_fmri_x=np.random.rand(64, 64, 35),
        ...     os_fmri_x=np.random.rand(64, 64, 60),
        ...     sharedmem=True,
        ...     offsettime=0.5,
        ...     ampthresh=0.5,
        ...     lagminthresh=0.3,
        ...     lagmaxthresh=2.0,
        ...     sigmathresh=500.0,
        ...     cleanrefined=True,
        ...     bipolar=False,
        ...     fixdelay=False,
        ...     includemask=None,
        ...     excludemask=None,
        ...     LGR=None,
        ...     nprocs=4,
        ...     detrendorder=2,
        ...     alwaysmultiproc=False,
        ...     showprogressbar=True,
        ...     chunksize=10000,
        ...     padtrs=5,
        ...     refineprenorm="mean",
        ...     refineweighting=None,
        ...     refinetype="pca",
        ...     pcacomponents=0.9,
        ...     dodispersioncalc=True,
        ...     dispersioncalc_lower=-4.0,
        ...     dispersioncalc_upper=4.0,
        ...     dispersioncalc_step=0.25,
        ...     windowfunc="hann",
        ...     passes=2,
        ...     maxpasses=10,
        ...     convergencethresh=None,
        ...     interptype="multivariate",
        ...     usetmask=True,
        ...     tmask_y=np.ones((64, 64)),
        ...     tmaskos_y=np.ones((64, 64)),
        ...     fastresamplerpadtime=30.0,
        ...     prewhitenregressor=True,
        ...     prewhitenlags=5,
        ...     debug=False,
        ...     rt_floattype=np.float32,
        ... )
        """
        self.internalvalidfmrishape = internalvalidfmrishape
        self.internalvalidpaddedfmrishape = internalvalidpaddedfmrishape
        self.sharedmem = sharedmem
        self.outputname = outputname
        self.initial_fmri_x = initial_fmri_x
        self.paddedinitial_fmri_x = paddedinitial_fmri_x
        self.os_fmri_x = os_fmri_x

        self.offsettime = offsettime
        self.ampthresh = ampthresh
        self.lagminthresh = lagminthresh
        self.lagmaxthresh = lagmaxthresh
        self.sigmathresh = sigmathresh
        self.cleanrefined = cleanrefined
        self.bipolar = bipolar
        self.fixdelay = fixdelay
        self.LGR = LGR
        self.nprocs = nprocs
        self.detrendorder = detrendorder
        self.alwaysmultiproc = alwaysmultiproc
        self.showprogressbar = showprogressbar
        self.chunksize = chunksize
        self.padtrs = padtrs
        self.refineprenorm = refineprenorm
        self.refineweighting = refineweighting
        self.refinetype = refinetype
        self.pcacomponents = pcacomponents
        self.dodispersioncalc = dodispersioncalc
        self.dispersioncalc_lower = dispersioncalc_lower
        self.dispersioncalc_upper = dispersioncalc_upper
        self.dispersioncalc_step = dispersioncalc_step
        self.windowfunc = windowfunc
        self.passes = passes
        self.maxpasses = maxpasses
        self.convergencethresh = convergencethresh
        self.interptype = interptype
        self.usetmask = usetmask
        self.tmask_y = tmask_y
        self.tmaskos_y = tmaskos_y
        self.fastresamplerpadtime = fastresamplerpadtime
        self.prewhitenregressor = prewhitenregressor
        self.prewhitenlags = prewhitenlags
        self.debug = debug
        self.rt_floattype = rt_floattype

        self.setmasks(includemask, excludemask)
        self.totalrefinementbytes = self._allocatemem(pid)

    def setmasks(self, includemask: Any, excludemask: Any) -> None:
        """
        Set the include and exclude masks for the object.

        Parameters
        ----------
        includemask : Any
            The mask to be used for including elements. Type and structure depends
            on the specific implementation and usage context.
        excludemask : Any
            The mask to be used for excluding elements. Type and structure depends
            on the specific implementation and usage context.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        This method assigns the provided masks to instance attributes `includemask`
        and `excludemask`. The masks are typically used for filtering or selection
        operations in data processing workflows.

        Examples
        --------
        >>> obj = MyClass()
        >>> obj.setmasks([1, 0, 1], [0, 1, 0])
        >>> print(obj.includemask)
        [1, 0, 1]
        >>> print(obj.excludemask)
        [0, 1, 0]
        """
        self.includemask = includemask
        self.excludemask = excludemask

    def _allocatemem(self, pid: Any) -> int:
        """
        Allocate memory for refinement arrays using shared memory if specified.

        This function allocates four arrays used in the refinement process:
        `shiftedtcs`, `weights`, `paddedshiftedtcs`, and `paddedweights`. These
        arrays are allocated with shapes determined by `internalvalidfmrishape` and
        `internalvalidpaddedfmrishape`, using the specified data type and memory
        sharing settings.

        Parameters
        ----------
        pid : Any
            Process identifier used to name shared memory segments.

        Returns
        -------
        int
            Total number of bytes allocated for the refinement arrays.

        Notes
        -----
        If `sharedmem` is True, the arrays are allocated in shared memory; otherwise,
        they are allocated locally. The function prints information about the
        allocation size and location, and logs memory usage after allocation.

        Examples
        --------
        >>> self._allocatemem(pid=1234)
        allocated 10.500 MB in shared memory for refinement
        11010048
        """
        self.shiftedtcs, self.shiftedtcs_shm = tide_util.allocarray(
            self.internalvalidfmrishape,
            self.rt_floattype,
            shared=self.sharedmem,
            name=f"shiftedtcs_{pid}",
        )
        self.weights, self.weights_shm = tide_util.allocarray(
            self.internalvalidfmrishape,
            self.rt_floattype,
            shared=self.sharedmem,
            name=f"weights_{pid}",
        )
        self.paddedshiftedtcs, self.paddedshiftedtcs_shm = tide_util.allocarray(
            self.internalvalidpaddedfmrishape,
            self.rt_floattype,
            shared=self.sharedmem,
            name=f"paddedshiftedtcs_{pid}",
        )
        self.paddedweights, self.paddedweights_shm = tide_util.allocarray(
            self.internalvalidpaddedfmrishape,
            self.rt_floattype,
            shared=self.sharedmem,
            name=f"paddedweights_{pid}",
        )
        if self.sharedmem:
            ramlocation = "in shared memory"
        else:
            ramlocation = "locally"
        totalrefinementbytes = (
            self.shiftedtcs.nbytes
            + self.weights.nbytes
            + self.paddedshiftedtcs.nbytes
            + self.paddedweights.nbytes
        )
        thesize, theunit = tide_util.format_bytes(totalrefinementbytes)
        print(f"allocated {thesize:.3f} {theunit} {ramlocation} for refinement")
        tide_util.logmem("after refinement array allocation")
        return totalrefinementbytes

    def cleanup(self) -> None:
        """
        Clean up memory resources by deleting internal attributes and shared memory segments.

        This method removes all internal arrays and their corresponding shared memory segments
        when shared memory is being used. It's designed to free up memory resources that were
        allocated during processing.

        Returns
        -------
        None
            This method does not return any value.

        Notes
        -----
        - Deletes the following internal attributes: paddedshiftedtcs, paddedweights,
          shiftedtcs, weights
        - If shared memory is enabled (sharedmem=True), also cleans up the corresponding
          shared memory segments using tide_util.cleanup_shm()
        - This method should be called when the object is no longer needed to prevent
          memory leaks

        Examples
        --------
        >>> obj = MyClass()
        >>> obj.cleanup()
        >>> # All internal memory resources are now freed
        """
        del self.paddedshiftedtcs
        del self.paddedweights
        del self.shiftedtcs
        del self.weights
        if self.sharedmem:
            tide_util.cleanup_shm(self.paddedshiftedtcs_shm)
            tide_util.cleanup_shm(self.paddedweights_shm)
            tide_util.cleanup_shm(self.shiftedtcs_shm)
            tide_util.cleanup_shm(self.weights_shm)

    def makemask(self, lagstrengths: Any, lagtimes: Any, lagsigma: Any, fitmask: Any) -> bool:
        """
        Create a refinement mask based on lag strength, lag time, and sigma thresholds.

        This function generates a mask for refining regressor parameters by evaluating
        the quality of lag estimates against specified thresholds. The mask determines
        which voxels should be included in the refinement process based on their
        lag strength, lag time, and sigma values.

        Parameters
        ----------
        lagstrengths : array-like
            Array containing lag strength values for each voxel
        lagtimes : array-like
            Array containing lag time values for each voxel
        lagsigma : array-like
            Array containing sigma values for each voxel
        fitmask : array-like
            Boolean mask indicating which voxels to consider for fitting

        Returns
        -------
        bool
            True if voxels are included in the refine mask, False if no voxels
            meet the refinement criteria

        Notes
        -----
        The function uses internal threshold parameters to determine which voxels
        should be included in the refinement process. These include:

        - ampthresh: amplitude threshold
        - lagminthresh: minimum lag threshold
        - lagmaxthresh: maximum lag threshold
        - sigmathresh: sigma threshold

        If no voxels meet the criteria, a critical log message is generated and
        the function returns False.

        Examples
        --------
        >>> # Assuming self is an instance of a class with the required attributes
        >>> result = self.makemask(lagstrengths, lagtimes, lagsigma, fitmask)
        >>> if result:
        ...     print("Refinement mask created successfully")
        ... else:
        ...     print("No voxels in refine mask")
        """
        # create the refinement mask
        (
            self.refinemaskvoxels,
            self.refinemask,
            self.locationfails,
            self.ampfails,
            self.lagfails,
            self.sigmafails,
            self.numinmask,
        ) = tide_refineregressor.makerefinemask(
            lagstrengths,
            lagtimes,
            lagsigma,
            fitmask,
            offsettime=self.offsettime,
            ampthresh=self.ampthresh,
            lagminthresh=self.lagminthresh,
            lagmaxthresh=self.lagmaxthresh,
            sigmathresh=self.sigmathresh,
            cleanrefined=self.cleanrefined,
            bipolar=self.bipolar,
            includemask=self.includemask,
            excludemask=self.excludemask,
            fixdelay=self.fixdelay,
            debug=self.debug,
        )

        if self.numinmask == 0:
            self.LGR.critical("No voxels in refine mask - adjust thresholds or external masks")
            return False
        else:
            return True

    def getrefinemask(self) -> NDArray:
        """
        Return the refinement mask.

        Returns
        -------
        NDArray
            The refinement mask stored in the instance.

        Notes
        -----
        This method provides access to the refinement mask attribute. The refinement
        mask is typically used in computational physics or data analysis workflows
        to identify regions of interest or apply specific processing to certain data
        points.

        Examples
        --------
        >>> mask = obj.getrefinemask()
        >>> print(mask)
        [True, False, True, False]
        """
        return self.refinemask

    def getpaddedshiftedtcs(self) -> NDArray:
        """
        Return the padded and shifted time-course data.

        This method retrieves the pre-computed padded and shifted time-course data
        that has been processed for analysis. The data is typically used in
        time-series analysis or signal processing applications where temporal
        alignment and padding are required.

        Returns
        -------
        NDArray
            The padded and shifted time-course data stored in the instance variable
            `paddedshiftedtcs`. The exact format depends on the data processing
            pipeline that generated this data.

        Notes
        -----
        This method serves as a simple getter function for the `paddedshiftedtcs`
        attribute. The actual computation of padded and shifted time-course data
        should be performed prior to calling this method.

        Examples
        --------
        >>> processor = TimeCourseProcessor()
        >>> processor.compute_padded_shifted_tcs()
        >>> result = processor.getpaddedshiftedtcs()
        >>> print(result.shape)
        (1000, 50)
        """
        return self.paddedshiftedtcs

    def alignvoxels(self, fmri_data_valid: Any, fmritr: Any, lagtimes: Any) -> int:
        """
        Align timecourses to prepare for refinement.

        This function aligns voxel timecourses for further refinement processing by
        coordinating the alignment of fMRI data with specified lag times and processing
        parameters.

        Parameters
        ----------
        fmri_data_valid : Any
            Valid fMRI data to be aligned
        fmritr : Any
            fMRI temporal resolution information
        lagtimes : Any
            Lag times to be used for alignment

        Returns
        -------
        int
            Returns the number of voxels processed

        Notes
        -----
        The function utilizes the `tide_refineregressor.alignvoxels` function internally
        and passes all relevant processing parameters including multiprocessing settings,
        detrending options, and padding parameters.

        Examples
        --------
        >>> alignvoxels(fmri_data_valid, fmritr, lagtimes)
        >>> # Function processes data and updates internal state
        """
        # align timecourses to prepare for refinement
        self.LGR.info("aligning timecourses")
        voxelsprocessed_rra = tide_refineregressor.alignvoxels(
            fmri_data_valid,
            fmritr,
            self.shiftedtcs,
            self.weights,
            self.paddedshiftedtcs,
            self.paddedweights,
            lagtimes,
            self.refinemask,
            nprocs=self.nprocs,
            detrendorder=self.detrendorder,
            offsettime=self.offsettime,
            alwaysmultiproc=self.alwaysmultiproc,
            showprogressbar=self.showprogressbar,
            chunksize=self.chunksize,
            padtrs=self.padtrs,
            rt_floattype=self.rt_floattype,
        )
        return voxelsprocessed_rra
        # self.LGR.info(f"align complete: {voxelsprocessed_rra=}")

    def prenormalize(self, lagtimes: Any, lagstrengths: Any, R2: Any) -> None:
        """
        Pre-normalize time series data for refinement regression.

        This function applies pre-normalization to the padded and shifted time series
        data using the specified lag times, lag strengths, and R2 values. The
        normalization is performed through the underlying tide_refineregressor.prenorm
        function with the appropriate internal parameters.

        Parameters
        ----------
        lagtimes : Any
            Array or list of lag times to be used in the pre-normalization process.
        lagstrengths : Any
            Array or list of lag strengths corresponding to the lag times.
        R2 : Any
            Array or list of R2 values used for the pre-normalization calculation.

        Returns
        -------
        None
            This function does not return any value. It modifies internal attributes
            in-place.

        Notes
        -----
        The function internally uses:
        - self.paddedshiftedtcs: padded and shifted time series data
        - self.refinemask: refinement mask for the normalization process
        - self.lagmaxthresh: maximum lag threshold
        - self.refineprenorm: refinement pre-normalization parameters
        - self.refineweighting: refinement weighting parameters

        Examples
        --------
        >>> prenormalize(lagtimes=[1, 2, 3], lagstrengths=[0.5, 0.3, 0.8], R2=[0.9, 0.85, 0.92])
        """
        tide_refineregressor.prenorm(
            self.paddedshiftedtcs,
            self.refinemask,
            lagtimes,
            self.lagmaxthresh,
            lagstrengths,
            R2,
            self.refineprenorm,
            self.refineweighting,
        )

    def refine(
        self,
        theprefilter: Any,
        fmritr: Any,
        thepass: Any,
        lagstrengths: Any,
        lagtimes: Any,
        previousnormoutputdata: Any,
        corrmasksize: Any,
    ) -> Tuple[int, dict, NDArray, NDArray, NDArray, bool, Optional[str], Any]:
        """
        Refine the regressor by iteratively applying filtering and resampling operations.

        This method performs iterative refinement of a regressor using a series of
        filtering, resampling, and normalization steps. It tracks convergence and
        updates internal state variables accordingly.

        Parameters
        ----------
        theprefilter : Any
            The prefilter to be applied to the data.
        fmritr : Any
            The fMRI temporal resolution (TR).
        thepass : Any
            The current pass number in the refinement process.
        lagstrengths : Any
            The lag strengths used in the refinement.
        lagtimes : Any
            The lag times used in the refinement.
        previousnormoutputdata : Any
            The normalized output data from the previous pass.
        corrmasksize : Any
            The size of the correlation mask.

        Returns
        -------
        tuple
            A tuple containing:
            - voxelsprocessed_rr : int
                Number of voxels processed in this pass.
            - outputdict : dict
                Dictionary of output statistics for this pass.
            - previousnormoutputdata : NDArray
                Updated normalized output data for the next pass.
            - resampref_y : NDArray
                Resampled refined regressor at oversampled frequency.
            - resampnonosref_y : NDArray
                Resampled refined regressor at original frequency.
            - stoprefining : bool
                Flag indicating whether refinement should stop.
            - refinestopreason : str or None
                Reason for stopping refinement.
            - genlagtc : FastResampler
                Generator for lag time courses.

        Notes
        -----
        This function modifies internal attributes of the class, such as `paddedoutputdata`,
        `locationfails`, `ampfails`, `lagfails`, and `sigmafails`. It also writes output
        files using `tide_io.writebidstsv`.

        Examples
        --------
        >>> refine(prefilter, fmritr, 1, lagstrengths, lagtimes, prev_data, mask_size)
        """
        (
            voxelsprocessed_rr,
            self.paddedoutputdata,
        ) = tide_refineregressor.dorefine(
            self.paddedshiftedtcs,
            self.refinemask,
            self.weights,
            theprefilter,
            fmritr,
            thepass,
            lagstrengths,
            lagtimes,
            self.refinetype,
            1.0 / fmritr,
            self.outputname,
            detrendorder=self.detrendorder,
            pcacomponents=self.pcacomponents,
            dodispersioncalc=self.dodispersioncalc,
            dispersioncalc_lower=self.dispersioncalc_lower,
            dispersioncalc_upper=self.dispersioncalc_upper,
            dispersioncalc_step=self.dispersioncalc_step,
            windowfunc=self.windowfunc,
            cleanrefined=self.cleanrefined,
            bipolar=self.bipolar,
            rt_floattype=self.rt_floattype,
        )
        outputdict = {}
        outputdict["refinemasksize_pass" + str(thepass)] = voxelsprocessed_rr
        outputdict["refinemaskpct_pass" + str(thepass)] = 100.0 * voxelsprocessed_rr / corrmasksize
        outputdict["refinelocationfails_pass" + str(thepass)] = self.locationfails
        outputdict["refineampfails_pass" + str(thepass)] = self.ampfails
        outputdict["refinelagfails_pass" + str(thepass)] = self.lagfails
        outputdict["refinesigmafails_pass" + str(thepass)] = self.sigmafails

        if self.prewhitenregressor:
            self.paddedoutputdata = tide_fit.prewhiten(
                self.paddedoutputdata, self.prewhitenlags, debug=self.debug
            )

        fmrifreq = 1.0 / fmritr
        if voxelsprocessed_rr > 0:
            paddednormoutputdata = tide_math.stdnormalize(
                theprefilter.apply(fmrifreq, self.paddedoutputdata)
            )
            outputdata = self.paddedoutputdata[self.padtrs : -self.padtrs]
            normoutputdata = tide_math.stdnormalize(theprefilter.apply(fmrifreq, outputdata))
            normunfilteredoutputdata = tide_math.stdnormalize(outputdata)
            tide_io.writebidstsv(
                f"{self.outputname}_desc-refinedmovingregressor_timeseries",
                normunfilteredoutputdata,
                fmrifreq,
                columns=["unfiltered_pass" + str(thepass)],
                extraheaderinfo={
                    "Description": "The raw and filtered probe regressor produced by the refinement procedure, at the time resolution of the data"
                },
                append=(thepass > 1),
            )
            tide_io.writebidstsv(
                f"{self.outputname}_desc-refinedmovingregressor_timeseries",
                normoutputdata,
                fmrifreq,
                columns=["filtered_pass" + str(thepass)],
                extraheaderinfo={
                    "Description": "The raw and filtered probe regressor produced by the refinement procedure, at the time resolution of the data"
                },
                append=True,
            )

        # check for convergence
        regressormse = mse(normoutputdata, previousnormoutputdata)
        outputdict["regressormse_pass" + str(thepass).zfill(2)] = regressormse
        self.LGR.info(f"regressor difference at end of pass {thepass:d} is {regressormse:.6f}")
        if self.convergencethresh is not None:
            if thepass >= self.maxpasses:
                self.LGR.info("refinement ended (maxpasses reached)")
                stoprefining = True
                refinestopreason = "maxpassesreached"
            elif regressormse < self.convergencethresh:
                self.LGR.info("refinement ended (refinement has converged")
                stoprefining = True
                refinestopreason = "convergence"
            else:
                stoprefining = False
        elif thepass >= self.passes:
            stoprefining = True
            refinestopreason = "passesreached"
        else:
            stoprefining = False
            refinestopreason = None
        outputdict["refinestopreason"] = refinestopreason

        if self.detrendorder > 0:
            resampnonosref_y = tide_fit.detrend(
                tide_resample.doresample(
                    self.paddedinitial_fmri_x,
                    paddednormoutputdata,
                    self.initial_fmri_x,
                    method=self.interptype,
                ),
                order=self.detrendorder,
                demean=True,
            )
            resampref_y = tide_fit.detrend(
                tide_resample.doresample(
                    self.paddedinitial_fmri_x,
                    paddednormoutputdata,
                    self.os_fmri_x,
                    method=self.interptype,
                ),
                order=self.detrendorder,
                demean=True,
            )
        else:
            resampnonosref_y = tide_resample.doresample(
                self.paddedinitial_fmri_x,
                paddednormoutputdata,
                self.initial_fmri_x,
                method=self.interptype,
            )
            resampref_y = tide_resample.doresample(
                self.paddedinitial_fmri_x,
                paddednormoutputdata,
                self.os_fmri_x,
                method=self.interptype,
            )
        if self.usetmask:
            resampnonosref_y *= self.tmask_y
            thefit, R2val = tide_fit.mlregress(self.tmask_y, resampnonosref_y)
            resampnonosref_y -= thefit[0, 1] * self.tmask_y
            resampref_y *= self.tmaskos_y
            thefit, R2val = tide_fit.mlregress(self.tmaskos_y, resampref_y)
            resampref_y -= thefit[0, 1] * self.tmaskos_y

        # reinitialize genlagtc for resampling
        previousnormoutputdata = np.zeros_like(normoutputdata)
        genlagtc = tide_resample.FastResampler(
            self.paddedinitial_fmri_x,
            paddednormoutputdata,
            padtime=self.fastresamplerpadtime,
        )
        genlagtc.save(f"{self.outputname}_desc-lagtcgenerator_timeseries")
        if self.debug:
            genlagtc.info()
        (
            outputdict[f"kurtosis_reference_pass{thepass + 1}"],
            outputdict[f"kurtosisz_reference_pass{thepass + 1}"],
            outputdict[f"kurtosisp_reference_pass{thepass + 1}"],
        ) = tide_stats.kurtosisstats(resampref_y)
        (
            outputdict[f"skewness_reference_pass{thepass + 1}"],
            outputdict[f"skewnessz_reference_pass{thepass + 1}"],
            outputdict[f"skewnessp_reference_pass{thepass + 1}"],
        ) = tide_stats.skewnessstats(resampref_y)
        if not stoprefining:
            tide_io.writebidstsv(
                f"{self.outputname}_desc-movingregressor_timeseries",
                tide_math.stdnormalize(resampnonosref_y),
                1.0 / fmritr,
                columns=["pass" + str(thepass + 1)],
                extraheaderinfo={
                    "Description": "The probe regressor used in each pass, at the time resolution of the data"
                },
                append=True,
            )
            oversampfreq = 1.0 / (self.os_fmri_x[1] - self.os_fmri_x[0])
            tide_io.writebidstsv(
                f"{self.outputname}_desc-oversampledmovingregressor_timeseries",
                tide_math.stdnormalize(resampref_y),
                oversampfreq,
                columns=["pass" + str(thepass + 1)],
                extraheaderinfo={
                    "Description": "The probe regressor used in each pass, at the time resolution used for calculating the similarity function"
                },
                append=True,
            )
        else:
            self.LGR.warning(f"refinement failed - terminating at end of pass {thepass}")
            stoprefining = True
            refinestopreason = "emptymask"

        return (
            voxelsprocessed_rr,
            outputdict,
            previousnormoutputdata,
            resampref_y,
            resampnonosref_y,
            stoprefining,
            refinestopreason,
            genlagtc,
        )


def refineRegressor(
    LGR: Any,
    TimingLGR: Any,
    thepass: Any,
    optiondict: Any,
    fitmask: NDArray,
    internaloffsetincludemask_valid: NDArray,
    internaloffsetexcludemask_valid: NDArray,
    internalrefineincludemask_valid: NDArray,
    internalrefineexcludemask_valid: NDArray,
    internaldespeckleincludemask: NDArray,
    validvoxels: NDArray,
    theRegressorRefiner: Any,
    lagtimes: NDArray,
    lagstrengths: NDArray,
    lagsigma: NDArray,
    fmri_data_valid: NDArray,
    fmritr: float,
    R2: Any,
    theprefilter: Any,
    previousnormoutputdata: Any,
    theinputdata: Any,
    numpadtrs: Any,
    outputname: Any,
    nativefmrishape: Any,
    bidsbasedict: Any,
    rt_floattype: np.dtype = np.dtype(np.float64),
    debug: bool = False,
) -> Tuple[NDArray, NDArray, bool, str, Any]:
    """
    Refine the regressor by adjusting masks, aligning timecourses, and performing refinement steps.

    This function performs regressor refinement during a specified pass, including:
    - Updating offset time based on lag properties
    - Managing masks for refinement and despeckling
    - Aligning timecourses
    - Pre-normalizing data
    - Executing the refinement step using a regressor refiner object

    Parameters
    ----------
    LGR : logging.Logger
        Logger instance for general logging.
    TimingLGR : logging.Logger
        Logger instance for timing-related messages.
    thepass : int
        Current pass number in the refinement process.
    optiondict : dict
        Dictionary containing various options and settings for the refinement process.
    fitmask : array_like
        Boolean mask indicating voxels to be considered in fitting.
    internaloffsetincludemask_valid : array_like or None
        Mask for including voxels in offset calculation.
    internaloffsetexcludemask_valid : array_like or None
        Mask for excluding voxels from offset calculation.
    internalrefineincludemask_valid : array_like or None
        Mask for including voxels in refinement.
    internalrefineexcludemask_valid : array_like or None
        Mask for excluding voxels from refinement.
    internaldespeckleincludemask : array_like
        Mask for including voxels in despeckling.
    validvoxels : array_like
        Indices of valid voxels.
    theRegressorRefiner : object
        Regressor refiner object with methods for mask setting, masking, alignment, etc.
    lagtimes : array_like
        Array of lag times.
    lagstrengths : array_like
        Array of lag strengths.
    lagsigma : array_like
        Array of lag sigma values.
    fmri_data_valid : array_like
        Valid fMRI data.
    fmritr : float
        fMRI repetition time.
    R2 : array_like
        R2 values.
    theprefilter : object
        Filter object used for preprocessing.
    previousnormoutputdata : array_like
        Previously normalized output data.
    theinputdata : object
        Input data object (e.g., Nifti1Image).
    numpadtrs : int
        Number of padded timepoints.
    outputname : str
        Base name for output files.
    nativefmrishape : tuple
        Shape of the native fMRI data.
    bidsbasedict : dict
        Dictionary for BIDS metadata.
    rt_floattype : numpy.dtype, optional
        Data type for floating-point operations, default is np.float64.
    debug : bool, optional
        Enable debug mode, default is False.

    Returns
    -------
    tuple
        A tuple containing:
        - resampref_y : NDArray
          Resampled refined y values.
        - resampnonosref_y : NDArray
          Resampled non-oscillatory refined y values.
        - stoprefining : bool
          Flag indicating whether refinement should stop.
        - refinestopreason : str
          Reason for stopping refinement.
        - genlagtc : array_like
          Generated lag timecourses.

    Notes
    -----
    - This function modifies `optiondict` in-place, updating offset times and other parameters.
    - The function uses `theRegressorRefiner` to perform various refinement steps including:
      `setmasks`, `makemask`, `alignvoxels`, `prenormalize`, and `refine`.
    - If `refinedespeckled` is False and `despeckle_passes` > 0, the exclude mask for refinement
      is adjusted to include voxels being despeckled.
    - The function may exit early if no voxels qualify for refinement.

    Examples
    --------
    >>> refineRegressor(
    ...     LGR, TimingLGR, 1, optiondict, fitmask, offset_incl, offset_excl,
    ...     refine_incl, refine_excl, despeckle_incl, valid_voxels, refiner,
    ...     lagtimes, lagstrengths, lagsigma, fmri_data, fmritr, R2, prefilter,
    ...     prev_norm_data, input_data, numpadtrs, output_name, native_shape,
    ...     bids_dict, debug=True
    ... )
    """
    LGR.info(f"\n\nRegressor refinement, pass {thepass}")
    TimingLGR.info(f"Regressor refinement start, pass {thepass}")
    if optiondict["refineoffset"]:
        # check that we won't end up excluding all voxels from offset calculation before accepting mask
        offsetmask = np.uint16(fitmask)
        if internaloffsetincludemask_valid is not None:
            offsetmask[np.where(internaloffsetincludemask_valid == 0)] = np.uint16(0)
        if internaloffsetexcludemask_valid is not None:
            offsetmask[np.where(internaloffsetexcludemask_valid != 0.0)] = np.uint16(0)
        if tide_stats.getmasksize(offsetmask) == 0:
            LGR.warning(
                "NB: cannot exclude voxels from offset calculation mask - including for this pass"
            )
            offsetmask = fitmask + 0

        peaklag, dummy, dummy = tide_stats.gethistprops(
            lagtimes[np.where(offsetmask > 0)],
            optiondict["histlen"],
            pickleft=optiondict["pickleft"],
            peakthresh=optiondict["pickleftthresh"],
        )
        optiondict["offsettime"] = peaklag
        optiondict["offsettime_total"] += peaklag
        optiondict[f"offsettime_pass{thepass}"] = optiondict["offsettime"]
        optiondict[f"offsettime_total_pass{thepass}"] = optiondict["offsettime_total"]
        LGR.info(
            f"offset time set to {optiondict['offsettime']:.3f}, "
            f"total is {optiondict['offsettime_total']:.3f}"
        )

    if optiondict["refinedespeckled"] or (optiondict["despeckle_passes"] == 0):
        # if refinedespeckled is true, or there is no despeckling, masks are unaffected
        thisinternalrefineexcludemask_valid = internalrefineexcludemask_valid
    else:
        # if refinedespeckled is false and there is despeckling, need to make a proper mask
        if internalrefineexcludemask_valid is None:
            # if there is currently no exclude mask, set exclude mask = despeckle mask
            thisinternalrefineexcludemask_valid = np.where(
                internaldespeckleincludemask[validvoxels] == 0.0, 0, 1
            )
        else:
            # if there is a current exclude mask, add any voxels that are being despeckled
            thisinternalrefineexcludemask_valid = np.where(
                internalrefineexcludemask_valid > 0, 1, 0
            )
            thisinternalrefineexcludemask_valid[
                np.where(internaldespeckleincludemask[validvoxels] != 0.0)
            ] = 1

        # now check that we won't end up excluding all voxels from refinement before accepting mask
        overallmask = np.uint16(fitmask)
        if internalrefineincludemask_valid is not None:
            overallmask[np.where(internalrefineincludemask_valid == 0)] = np.uint16(0)
        if thisinternalrefineexcludemask_valid is not None:
            overallmask[np.where(thisinternalrefineexcludemask_valid != 0.0)] = np.uint16(0)
        if tide_stats.getmasksize(overallmask) == 0:
            LGR.warning(
                "NB: cannot exclude despeckled voxels from refinement - including for this pass"
            )
            thisinternalrefineexcludemask_valid = internalrefineexcludemask_valid
    theRegressorRefiner.setmasks(
        internalrefineincludemask_valid, thisinternalrefineexcludemask_valid
    )

    # regenerate regressor for next pass
    # create the refinement mask
    LGR.info("making refine mask")
    createdmask = theRegressorRefiner.makemask(lagstrengths, lagtimes, lagsigma, fitmask)
    print(f"Refine mask has {theRegressorRefiner.refinemaskvoxels} voxels")
    if not createdmask:
        print("no voxels qualify for refinement - exiting")
        sys.exit()

    # align timecourses to prepare for refinement
    LGR.info("aligning timecourses")
    tide_util.disablemkl(optiondict["nprocs_refine"], debug=optiondict["threaddebug"])
    voxelsprocessed_rra = theRegressorRefiner.alignvoxels(fmri_data_valid, fmritr, lagtimes)
    tide_util.enablemkl(optiondict["mklthreads"], debug=optiondict["threaddebug"])
    LGR.info(f"align complete: {voxelsprocessed_rra=}")

    # prenormalize
    LGR.info("prenormalizing timecourses")
    theRegressorRefiner.prenormalize(lagtimes, lagstrengths, R2)

    # now doing the refinement
    (
        voxelsprocessed_rr,
        outputdict,
        previousnormoutputdata,
        resampref_y,
        resampnonosref_y,
        stoprefining,
        refinestopreason,
        genlagtc,
    ) = theRegressorRefiner.refine(
        theprefilter,
        fmritr,
        thepass,
        lagstrengths,
        lagtimes,
        previousnormoutputdata,
        optiondict["corrmasksize"],
    )
    TimingLGR.info(
        f"Regressor refinement end, pass {thepass}",
        {
            "message2": voxelsprocessed_rr,
            "message3": "voxels",
        },
    )
    for key, value in outputdict.items():
        optiondict[key] = value

    # Save shifted timecourses for César
    if optiondict["saveintermediatemaps"] and optiondict["savelagregressors"]:
        theheader = theinputdata.copyheader()
        bidspasssuffix = f"_intermediatedata-pass{thepass}"
        maplist = [
            (
                (theRegressorRefiner.getpaddedshiftedtcs())[:, numpadtrs:-numpadtrs],
                "shiftedtcs",
                "bold",
                None,
                "The filtered input fMRI data, in voxels used for refinement, time shifted by the negated delay in every voxel so that the moving blood component is aligned.",
            ),
        ]
        tide_io.savemaplist(
            f"{outputname}{bidspasssuffix}",
            maplist,
            validvoxels,
            nativefmrishape,
            theheader,
            bidsbasedict,
            filetype=theinputdata.filetype,
            rt_floattype=rt_floattype,
            cifti_hdr=theinputdata.cifti_hdr,
            debug=debug,
        )

    return resampref_y, resampnonosref_y, stoprefining, refinestopreason, genlagtc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import os.path as op
import sys

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

import rapidtide.fit as tide_fit
import rapidtide.io as tide_io
import rapidtide.simFuncClasses as tide_simFuncClasses
from rapidtide.tests.utils import get_examples_path, get_test_temp_path


def dumplists(results, targets, failflags):
    print("assertion failed - dumping lists")
    if (len(results) != len(targets)) or (len(results) != len(failflags)):
        print("array lengths do not match")
        sys.exit()
    print("results", "targets")
    for i in range(len(results)):
        print(results[i], targets[i], failflags[i])


def eval_fml_result(
    absmin, absmax, testvalues, foundvalues, failflags, tolerance=0.0001, debug=False
):
    if debug:
        print(f"{absmin=}, {absmax=}, {tolerance=}")
        print(f"{testvalues=}")
        print(f"{foundvalues=}")
        print(f"{failflags=}")
    for i in range(len(testvalues)):
        if testvalues[i] < absmin:
            if foundvalues[i] != absmin:
                print(foundvalues[i], " != ", absmin, "for input", testvalues[i])
                dumplists(foundvalues, testvalues, failflags)
                return False
        elif testvalues[i] > absmax:
            if foundvalues[i] != absmax:
                print(foundvalues[i], " != ", absmax, "for input", testvalues[i])
                dumplists(foundvalues, testvalues, failflags)
                return False
        else:
            if np.fabs(foundvalues[i] - testvalues[i]) > tolerance:
                if failflags[i] == 0:
                    print("error found at index", i, failflags[i])
                    print(foundvalues[i], " != ", testvalues[i])
                    dumplists(foundvalues, testvalues, failflags)
                    return False
    return True


def test_findmaxlag(displayplots=False, local=False, debug=False):
    # set input and output directories
    exampleroot = get_examples_path(local)
    testtemproot = get_test_temp_path(local)

    if debug:
        print("debug flag is set")
    # for fittype in ["gauss", "quad", "fastquad", "COM", "None", "fastgauss", "gausscf"]:
    for fittype in ["gauss"]:
        print("*************************************")
        print(f"testing fittype: {fittype}")
        textfilename = op.join(exampleroot, "lt_rt.txt")

        # set default variable values
        searchfrac = 0.75

        indata = tide_io.readvecs(textfilename, debug=debug)
        xvecs = indata[0, :]
        yvecs = indata[1, :]

        # set some fit parameters
        lagmin = -20.0
        lagmax = 20.0
        widthmax = 1000.0
        absmaxsigma = 1000.0
        absminsigma = 0.1
        absmaxval = 1.0
        absminval = 0.0

        # test over the lag range
        testmaxval = 0.8
        testmaxsigma = 5.0
        testlags = np.linspace(-25.0, 25.0, 50, endpoint=True)
        testsigmas = np.full((len(testlags)), testmaxsigma, dtype=np.float64)
        testvals = np.full((len(testlags)), testmaxval, dtype=np.float64)

        fml_maxlags = np.zeros(len(testlags), dtype=np.float64)
        fml_maxsigmas = np.zeros(len(testlags), dtype=np.float64)
        fml_maxvals = np.zeros(len(testlags), dtype=np.float64)
        fml_lfailreasons = np.zeros(len(testlags), dtype=np.uint16)
        fmlc_maxlags = np.zeros(len(testlags), dtype=np.float64)
        fmlc_maxsigmas = np.zeros(len(testlags), dtype=np.float64)
        fmlc_maxvals = np.zeros(len(testlags), dtype=np.float64)
        fmlc_lfailreasons = np.zeros(len(testlags), dtype=np.uint16)

        # initialize the correlation fitter
        thefitter = tide_simFuncClasses.SimilarityFunctionFitter(
            corrtimeaxis=xvecs,
            lagmin=lagmin,
            lagmax=lagmax,
            absmaxsigma=absmaxsigma,
            absminsigma=absminsigma,
            peakfittype=fittype,
            debug=debug,
            searchfrac=searchfrac,
            zerooutbadfit=False,
        )

        for i in range(len(testlags)):
            yvecs = tide_fit.gauss_eval(xvecs, np.array([testvals[i], testlags[i], testsigmas[i]]))

            print()
            print()
            print()
            (
                maxindex,
                fml_maxlags[i],
                fml_maxvals[i],
                fml_maxsigmas[i],
                maskval,
                fml_lfailreasons[i],
                peakstart,
                peakend,
            ) = tide_fit.findmaxlag_gauss(
                xvecs,
                yvecs,
                lagmin,
                lagmax,
                widthmax,
                tweaklims=False,
                refine=True,
                debug=debug,
                searchfrac=searchfrac,
                absmaxsigma=absmaxsigma,
                absminsigma=absminsigma,
                zerooutbadfit=False,
            )

            print()
            print()
            print()
            (
                maxindexc,
                fmlc_maxlags[i],
                fmlc_maxvals[i],
                fmlc_maxsigmas[i],
                maskvalc,
                fmlc_lfailreasons[i],
                peakstartc,
                peakendc,
            ) = thefitter.fit(yvecs)
            print(
                maxindexc,
                fmlc_maxlags[i],
                fmlc_maxvals[i],
                fmlc_maxsigmas[i],
                maskvalc,
                fmlc_lfailreasons[i],
                peakstartc,
                peakendc,
            )

        if debug:
            print("findmaxlag_gauss results over lag range")
            for i in range(len(testlags)):
                print(testlags[i], fml_maxlags[i], fml_lfailreasons[i])

        assert eval_fml_result(
            lagmin, lagmax, testlags, fml_maxlags, fml_lfailreasons, debug=debug
        )
        assert eval_fml_result(
            absminval, absmaxval, testvals, fml_maxvals, fml_lfailreasons, debug=debug
        )
        assert eval_fml_result(
            absminsigma, absmaxsigma, testsigmas, fml_maxsigmas, fml_lfailreasons, debug=debug
        )

        assert eval_fml_result(
            lagmin, lagmax, testlags, fmlc_maxlags, fmlc_lfailreasons, debug=debug
        )
        assert eval_fml_result(
            absminval, absmaxval, testvals, fmlc_maxvals, fmlc_lfailreasons, debug=debug
        )
        assert eval_fml_result(
            absminsigma, absmaxsigma, testsigmas, fmlc_maxsigmas, fmlc_lfailreasons, debug=debug
        )

        if displayplots:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
            ax.plot(testlags, fml_maxlags, "r")
            ax.plot(testlags, fmlc_maxlags, "b")
            ax.legend(["findmaxlag_gauss", "classes"])
            plt.show()

        # now test over range of sigmas
        testlag = 5.0
        testsigmas = np.asarray(
            [
                0.01,
                0.02,
                0.05,
                0.1,
                0.2,
                0.5,
                1.0,
                2.0,
                5.0,
                10.0,
                20.0,
                50.0,
                100.0,
                200.0,
                500.0,
                1000.0,
                2000.0,
            ]
        )
        testlags = np.full((len(testsigmas)), testlag, dtype=np.float64)
        testvals = np.full((len(testsigmas)), testmaxval, dtype=np.float64)

        fml_maxlags = np.zeros(len(testsigmas), dtype=np.float64)
        fml_maxsigmas = np.zeros(len(testsigmas), dtype=np.float64)
        fml_maxvals = np.zeros(len(testsigmas), dtype=np.float64)
        fml_wfailreasons = np.zeros(len(testsigmas), dtype=np.uint16)
        fmlc_maxlags = np.zeros(len(testsigmas), dtype=np.float64)
        fmlc_maxsigmas = np.zeros(len(testsigmas), dtype=np.float64)
        fmlc_maxvals = np.zeros(len(testsigmas), dtype=np.float64)
        fmlc_wfailreasons = np.zeros(len(testsigmas), dtype=np.uint16)
        peakstartc = np.zeros(len(testsigmas), dtype=np.int32)
        peakendc = np.zeros(len(testsigmas), dtype=np.int32)

        for i in range(len(testsigmas)):
            yvecs = tide_fit.gauss_eval(xvecs, np.array([testvals[i], testlags[i], testsigmas[i]]))

            print()
            print()
            print()
            (
                maxindex,
                fml_maxlags[i],
                fml_maxvals[i],
                fml_maxsigmas[i],
                maskval,
                fml_wfailreasons[i],
                peakstart,
                peakend,
            ) = tide_fit.findmaxlag_gauss(
                xvecs,
                yvecs,
                lagmin,
                lagmax,
                widthmax,
                tweaklims=False,
                refine=True,
                debug=debug,
                searchfrac=searchfrac,
                absmaxsigma=absmaxsigma,
                absminsigma=absminsigma,
                zerooutbadfit=False,
            )

            print()
            print()
            print()
            (
                maxindexc,
                fmlc_maxlags[i],
                fmlc_maxvals[i],
                fmlc_maxsigmas[i],
                maskvalc,
                fmlc_wfailreasons[i],
                peakstartc[i],
                peakendc[i],
            ) = thefitter.fit(yvecs)
            print(
                maxindexc,
                fmlc_maxlags[i],
                fmlc_maxvals[i],
                fmlc_maxsigmas[i],
                maskvalc,
                fmlc_wfailreasons[i],
                peakstartc[i],
                peakendc[i],
            )

        if debug:
            print("findmaxlag_gauss results over sigma range")
            for i in range(len(testsigmas)):
                print(
                    testsigmas[i],
                    fml_maxsigmas[i],
                    fmlc_maxlags[i],
                    fmlc_maxvals[i],
                    fml_wfailreasons[i],
                )

            print("\nfitter class results over lag range")
            for i in range(len(testsigmas)):
                print(
                    testsigmas[i],
                    fmlc_maxsigmas[i],
                    fmlc_maxlags[i],
                    fmlc_maxvals[i],
                    peakstartc[i],
                    peakendc[i],
                    fmlc_wfailreasons[i],
                    thefitter.diagnosefail(np.uint32(fmlc_wfailreasons[i])),
                )

        if displayplots:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
            ax.loglog(testsigmas, fml_maxsigmas, "r")
            ax.loglog(testsigmas, fmlc_maxsigmas, "b")
            ax.legend(["findmaxlag_gauss", "classes"])
            plt.show()

        assert eval_fml_result(
            lagmin, lagmax, testlags, fml_maxlags, fml_wfailreasons, debug=debug
        )
        # assert eval_fml_result(absminval, absmaxval, testvals, fml_maxvals, fml_wfailreasons)
        assert eval_fml_result(
            absminsigma, absmaxsigma, testsigmas, fml_maxsigmas, fml_wfailreasons, debug=debug
        )

        assert eval_fml_result(
            lagmin, lagmax, testlags, fmlc_maxlags, fmlc_wfailreasons, debug=debug
        )
        assert eval_fml_result(
            absminval, absmaxval, testvals, fmlc_maxvals, fmlc_wfailreasons, debug=debug
        )
        assert eval_fml_result(
            absminsigma, absmaxsigma, testsigmas, fmlc_maxsigmas, fmlc_wfailreasons, debug=debug
        )


if __name__ == "__main__":
    mpl.use("TkAgg")
    test_findmaxlag(displayplots=True, local=True, debug=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import os
import sys
from typing import Any

import rapidtide.dlfiltertorch as tide_dlfilt
import rapidtide.happy_supportfuncs as happy_support
import rapidtide.io as tide_io
import rapidtide.miscmath as tide_math
import rapidtide.workflows.parser_funcs as pf

DEFAULT_MODEL = "model_cnn_pytorch"


def findfirst(searchlist: list, available: list, debug: bool = False) -> tuple[int, str | None]:
    if debug:
        print(f"FINDFIRST: {searchlist=}, {available=}")
    try:
        index = available.index(searchlist[0])
        return index, available[index]
    except ValueError:
        if len(searchlist) > 1:
            if debug:
                print("FINDFIRST: calling again")
            return findfirst(searchlist[1:], available, debug=debug)
        else:
            if debug:
                print("FINDFIRST: giving up")
    return -1, None


def _get_parser() -> Any:
    """
    Argument parser for applydlfilter.

    This function creates and configures an argument parser for the `applydlfilter`
    command-line tool, which applies a deep learning filter to a timecourse.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser object with defined arguments for `applydlfilter`.

    Notes
    -----
    The parser includes the following required and optional arguments:

    - ``infilename``: Input text file (or list of files) containing timecourse data.
    - ``outfilename``: Output text file (or list of files) to save filtered data.
    - ``--model``: Model root name to use for filtering (default: ``model_revised``).
    - ``--filesarelists``: Flag indicating input file contains lists of filenames.
    - ``--nodisplay``: Flag to disable plotting (for non-interactive use).
    - ``--verbose``: Flag to enable verbose output.

    Examples
    --------
    >>> parser = _get_parser()
    >>> args = parser.parse_args(['input.txt', 'output.txt', '--model', 'custom_model'])
    >>> print(args.infilename)
    'input.txt'
    >>> print(args.model)
    'custom_model'
    """
    parser = argparse.ArgumentParser(
        prog="applydlfilter",
        description=("Apply a deep learning filter to a timecourse."),
        allow_abbrev=False,
    )

    # Required arguments
    parser.add_argument(
        "infilename",
        type=lambda x: pf.is_valid_file(parser, x),
        help="The name of the input text file (or a list of names of input files).",
    )
    parser.add_argument(
        "outfilename",
        help="The name of the output text file (or a list of names of output files).",
    )

    # add optional arguments
    parser.add_argument(
        "--model",
        dest="model",
        action="store",
        metavar="MODELROOT",
        type=str,
        help=(f"Use model named MODELROOT (default is {DEFAULT_MODEL})."),
        default=DEFAULT_MODEL,
    )
    parser.add_argument(
        "--filesarelists",
        dest="filesarelists",
        action="store_true",
        help=("Input file contains lists of filenames, rather than data."),
        default=False,
    )
    parser.add_argument(
        "--nodisplay",
        dest="display",
        action="store_false",
        help=("Do not plot the data (for noninteractive use)."),
        default=True,
    )
    parser.add_argument(
        "--verbose",
        dest="verbose",
        action="store_true",
        help=("Print a lot of internal information."),
        default=False,
    )
    return parser


def applydlfilter(args: Any) -> None:
    """
    Apply a deep learning filter to fMRI data files.

    This function reads fMRI data from input files, applies a deep learning filter
    to denoise the data, and writes the filtered output to specified files. It supports
    processing multiple files either from lists or a single file, and optionally displays
    the filtering results using matplotlib.

    Parameters
    ----------
    args : Any
        An object containing the following attributes:
        - `infilename` : str
            Path to the input file or list of input files.
        - `outfilename` : str
            Path to the output file or list of output files.
        - `filesarelists` : bool
            If True, `infilename` and `outfilename` are treated as paths to text files
            containing lists of input and output filenames, respectively.
        - `model` : str
            Path to the deep learning model to be used for filtering.
        - `display` : bool
            If True, displays the original and filtered data using matplotlib.
        - `verbose` : bool
            If True, prints verbose output during processing.

    Returns
    -------
    None
        This function does not return any value. It writes filtered data to files
        and optionally displays plots.

    Notes
    -----
    - The function assumes that the input data has a sampling rate of 25.0 Hz.
    - If `filesarelists` is True, the input and output filenames are read from
      text files, where each line contains a single filename.
    - The function checks for matching list lengths when processing multiple files.
    - If a bad points file is specified, it is read and used during filtering.
    - The deep learning model is loaded from a predefined model path within the
      `rapidtide` package.

    Examples
    --------
    >>> import argparse
    >>> args = argparse.Namespace(
    ...     infilename="input.txt",
    ...     outfilename="output.txt",
    ...     filesarelists=False,
    ...     model="model.h5",
    ...     display=False,
    ...     verbose=True
    ... )
    >>> applydlfilter(args)
    """
    if args.display:
        import matplotlib as mpl

        mpl.use("TkAgg")
        import matplotlib.pyplot as plt

    if args.filesarelists:
        infilenamelist = []
        with open(args.infilename, "r") as f:
            inputlist = f.readlines()
            for line in inputlist:
                infilenamelist.append(line.strip())
                if args.verbose:
                    print(infilenamelist[-1])
        outfilenamelist = []
        with open(args.outfilename, "r") as f:
            inputlist = f.readlines()
            for line in inputlist:
                outfilenamelist.append(line.strip())
                if args.verbose:
                    print(outfilenamelist[-1])
        if len(infilenamelist) != len(outfilenamelist):
            print("list lengths do not match - exiting")
            sys.exit()
    else:
        infilenamelist = [args.infilename]
        outfilenamelist = [args.outfilename]

    # load the filter
    modelpath = os.path.join(
        os.path.split(os.path.split(os.path.split(__file__)[0])[0])[0],
        "rapidtide",
        "data",
        "models",
    )
    thedlfilter = tide_dlfilt.DeepLearningFilter(modelpath=modelpath)
    thedlfilter.loadmodel(args.model)
    usebadpts = thedlfilter.usebadpts
    showpleth = True

    # set the list of column names
    datanames = [
        "cardiacfromfmri_25.0Hz",
        "normcardiac_25.0Hz",
        "normcardiacfromfmri_dlfiltered_25.0Hz",
        "cardiacfromfmri_dlfiltered_25.0Hz",
    ]
    badptsnames = ["badpts"]
    plethnames = [
        "pleth",
        "normpleth",
        "plethenv",
        "pleth_dlfiltered",
    ]

    for idx, infilename in enumerate(infilenamelist):
        # read in the data
        if args.verbose:
            print("reading in", infilename)
        (
            thesamplerate,
            thestarttime,
            thecolumns,
            thedata,
            compressed,
            columnsource,
            extrainfo,
        ) = tide_io.readbidstsv(infilename, debug=args.verbose)

        # check to see what we have in the file
        if args.verbose:
            print(f"File columns: {thecolumns}")

        # get the data
        datacol, dataname = findfirst(datanames, thecolumns, debug=args.verbose)
        if dataname is None:
            print("file contains no usable raw data")
            sys.exit()
        if args.verbose:
            print(f"Data column: {datacol}, ({dataname})")
        fmridata = thedata[datacol, :]

        # get the badpts
        badptscol, badptsname = findfirst(badptsnames, thecolumns, debug=args.verbose)
        if usebadpts and (badptsname is None):
            print("file contains no usable badpts data")
            sys.exit()
        if args.verbose:
            print(f"Bad points column: {badptscol}, ({badptsname})")
        if badptsname is None:
            badpts = None
        else:
            badpts = thedata[badptscol, :]

        # get the pleth (if it is there)
        plethcol, plethname = findfirst(plethnames, thecolumns, debug=args.verbose)
        if args.verbose:
            print(f"Pleth column: {plethcol}, ({plethname})")
            print(f"{thedata.shape=}")
        if plethname is None:
            plethwave = None
        else:
            plethwave = thedata[plethcol, :]

        if args.verbose:
            print("data is read")
        if thesamplerate != 25.0:
            print("sampling rate", thesamplerate)
            sys.exit()
        if args.verbose:
            print("filtering...")
        predicteddata = thedlfilter.apply(fmridata, badpts=badpts)
        if args.verbose:
            print("done...")

        # performance metrics
        extradict = {}
        maxval, maxdelay, failreason = happy_support.checkcardmatch(
            fmridata, predicteddata, 25.0, debug=False
        )
        print(infilename, "max correlation of raw data to dl filtered data:", maxval)
        extradict["corr_rawtodlfiltered"] = maxval + 0.0

        if plethwave is not None:
            maxval, maxdelay, failreason = happy_support.checkcardmatch(
                fmridata, plethwave, 25.0, debug=False
            )
            print(infilename, "max correlation of raw data to target plethysmogram:", maxval)
            extradict["corr_rawtopleth"] = maxval + 0.0

            maxval, maxdelay, failreason = happy_support.checkcardmatch(
                predicteddata, plethwave, 25.0, debug=False
            )
            print(
                infilename, "max correlation of dl filtered data to target plethysmogram:", maxval
            )
            extradict["corr_dlfilteredtopleth"] = maxval + 0.0

        if args.verbose:
            print("writing to", outfilenamelist[idx])
        tide_io.writebidstsv(
            outfilenamelist[idx],
            predicteddata,
            25.0,
            extraheaderinfo=extradict,
            columns=["filtered_signal"],
            debug=args.verbose,
        )

        # normalize
        fmridata = tide_math.stdnormalize(fmridata)
        predicteddata = tide_math.stdnormalize(predicteddata)

        spacing = 3.0
        numwaves = 2
        if badpts is not None:
            numwaves += 1
        if plethwave is not None:
            numwaves += 1
        offset = (numwaves - 1) * spacing
        if args.display:
            plt.figure()
            legendlist = [dataname, "filtereddata"]
            plt.plot(fmridata + offset)
            offset += spacing
            plt.plot(predicteddata + offset)
            offset += spacing
            if plethwave is not None:
                legendlist.append(plethname)
                plt.plot(plethwave + offset)
                offset += spacing
            if badpts is not None:
                legendlist.append(badptsname)
                plt.plot(badpts + offset)
                offset += spacing
            plt.legend(legendlist)
            plt.title(args.model)
            plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import os

import matplotlib as mpl
import numpy as np

import rapidtide.io as tide_io
import rapidtide.workflows.rapidtide as rapidtide_workflow
import rapidtide.workflows.rapidtide_parser as rapidtide_parser
import rapidtide.workflows.retroregress as rapidtide_retroregress
from rapidtide.tests.utils import get_examples_path, get_test_temp_path, mse


def test_fullrunrapidtide_v7(debug=False, local=False, displayplots=False):
    # set input and output directories
    exampleroot = get_examples_path(local)
    testtemproot = get_test_temp_path(local)

    # test anatomic masks
    inputargs = [
        os.path.join(exampleroot, "sub-RAPIDTIDETEST.nii.gz"),
        os.path.join(testtemproot, "sub-RAPIDTIDETEST_seg"),
        "--nprocs",
        "-1",
        "--passes",
        "3",
        "--brainmask",
        os.path.join(exampleroot, "sub-RAPIDTIDETEST_brainmask.nii.gz"),
        "--graymattermask",
        os.path.join(exampleroot, "sub-RAPIDTIDETEST_synthseg.nii.gz:SSEG_GRAY"),
        "--whitemattermask",
        os.path.join(exampleroot, "sub-RAPIDTIDETEST_synthseg.nii.gz:SSEG_WHITE"),
        "--csfmask",
        os.path.join(exampleroot, "sub-RAPIDTIDETEST_synthseg.nii.gz:SSEG_CSF"),
    ]
    rapidtide_workflow.rapidtide_main(rapidtide_parser.process_args(inputargs=inputargs))

    inputargs = [
        os.path.join(exampleroot, "sub-RAPIDTIDETEST.nii.gz"),
        os.path.join(testtemproot, "sub-RAPIDTIDETEST_seg"),
        "--alternateoutput",
        os.path.join(testtemproot, "segtest"),
        "--nprocs",
        "-1",
        "--outputlevel",
        "max",
    ]
    rapidtide_retroregress.retroregress(rapidtide_retroregress.process_args(inputargs=inputargs))

    inputargs = [
        os.path.join(exampleroot, "sub-RAPIDTIDETEST.nii.gz"),
        os.path.join(testtemproot, "sub-RAPIDTIDETEST_seg"),
        "--alternateoutput",
        os.path.join(testtemproot, "regressoronly"),
        "--nprocs",
        "-1",
        "--outputlevel",
        "onlyregressors",
    ]
    rapidtide_retroregress.retroregress(rapidtide_retroregress.process_args(inputargs=inputargs))

    # check to see that rapidtide and retroregress output match
    msethresh = 2e-6
    aethresh = 2
    tclist = ["brain", "GM", "WM", "CSF"]
    for timecourse in tclist:
        infilespec = os.path.join(
            testtemproot,
            f"sub-RAPIDTIDETEST_seg_desc-regionalpostfilter_timeseries.json:{timecourse}",
        )
        insamplerate, instarttime, incolumns, indata, incompressed, infiletype = (
            tide_io.readvectorsfromtextfile(infilespec, onecol=True, debug=debug)
        )
        outfilespec = os.path.join(
            testtemproot, f"segtest_desc-regionalpostfilter_timeseries.json:{timecourse}"
        )
        outsamplerate, outstarttime, outcolumns, outdata, outcompressed, outfiletype = (
            tide_io.readvectorsfromtextfile(outfilespec, onecol=True, debug=debug)
        )
        assert insamplerate == outsamplerate
        assert instarttime == outstarttime
        assert incompressed == outcompressed
        assert infiletype == outfiletype
        assert incolumns == outcolumns
        assert indata.shape == outdata.shape
        assert indata.shape[0] == indata.shape[0]
        assert mse(indata, outdata) < msethresh
        np.testing.assert_almost_equal(indata, outdata, aethresh)


if __name__ == "__main__":
    mpl.use("TkAgg")
    test_fullrunrapidtide_v7(debug=True, local=True, displayplots=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2026-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from rapidtide.tests.utils import create_dir, get_test_temp_path
from rapidtide.workflows.spatialmi import _get_parser, getMI, getneighborhood, spatialmi

# ==================== Helpers ====================


def _make_mock_hdr(xsize, ysize, numslices, timepoints=1):
    """Create a mock NIfTI header that supports __getitem__ and __setitem__."""
    store = {
        "dim": [4, xsize, ysize, numslices, timepoints, 1, 1, 1],
        "pixdim": [1.0, 2.0, 2.0, 2.0, 1.5, 1.0, 1.0, 1.0],
    }
    hdr = MagicMock()
    hdr.__getitem__ = MagicMock(side_effect=lambda key: store[key])
    hdr.__setitem__ = MagicMock(side_effect=lambda key, val: store.__setitem__(key, val))

    def copy_fn():
        new_store = {
            "dim": list(store["dim"]),
            "pixdim": list(store["pixdim"]),
        }
        h = MagicMock()
        h.__getitem__ = MagicMock(side_effect=lambda key: new_store[key])
        h.__setitem__ = MagicMock(side_effect=lambda key, val: new_store.__setitem__(key, val))
        return h

    hdr.copy = copy_fn
    return hdr


def _make_dims(xsize, ysize, numslices, timepoints):
    return np.array([4, xsize, ysize, numslices, timepoints, 1, 1, 1])


def _make_sizes():
    return np.array([1.0, 2.0, 2.0, 2.0, 1.5, 1.0, 1.0, 1.0])


def _reset_getneighborhood_globals():
    """Reset the global variables used by getneighborhood for caching."""
    import rapidtide.workflows.spatialmi as sm

    for gname in ("kernel", "usedwidth", "indexlist", "usedradius"):
        if gname in dir(sm):
            try:
                delattr(sm, gname)
            except AttributeError:
                pass
    # Also reset via globals dict in the module
    for gname in ("kernel", "usedwidth", "indexlist", "usedradius"):
        if gname in sm.__dict__:
            del sm.__dict__[gname]


def _reset_getMI_globals():
    """Reset the global thebins variable used by getMI."""
    import rapidtide.workflows.spatialmi as sm

    sm.thebins = None


def _make_default_args(outputroot="/tmp/test_spatialmi"):
    """Create a default Namespace with all required args."""
    return argparse.Namespace(
        inputfilename1="dummy_input1.nii.gz",
        maskfilename1="dummy_mask1.nii.gz",
        inputfilename2="dummy_input2.nii.gz",
        maskfilename2="dummy_mask2.nii.gz",
        outputroot=outputroot,
        radius=2.0,
        sigma=None,
        spherical=False,
        kernelwidth=None,
        norm=True,
        prebin=True,
        index1=0,
        index2=0,
        debug=False,
    )


# ==================== _get_parser tests ====================


def parser_basic(debug=False):
    """Test that parser creates successfully."""
    if debug:
        print("parser_basic")
    parser = _get_parser()
    assert parser is not None
    assert parser.prog == "spatialmi"


def parser_required_args(debug=False):
    """Test parser has required positional arguments."""
    if debug:
        print("parser_required_args")
    parser = _get_parser()
    actions = {a.dest: a for a in parser._actions}
    assert "inputfilename1" in actions
    assert "maskfilename1" in actions
    assert "inputfilename2" in actions
    assert "maskfilename2" in actions
    assert "outputroot" in actions


def parser_defaults(debug=False):
    """Test default values for optional arguments."""
    if debug:
        print("parser_defaults")
    parser = _get_parser()
    args = parser.parse_args(["in1.nii", "mask1.nii", "in2.nii", "mask2.nii", "outroot"])
    assert args.prebin is True
    assert args.norm is True
    assert args.radius == 2.0
    assert args.sigma is None
    assert args.kernelwidth is None
    assert args.spherical is False
    assert args.index1 == 0
    assert args.index2 == 0
    assert args.debug is False


def parser_optional_args(debug=False):
    """Test optional arguments parsing."""
    if debug:
        print("parser_optional_args")
    parser = _get_parser()
    args = parser.parse_args(
        [
            "in1.nii",
            "mask1.nii",
            "in2.nii",
            "mask2.nii",
            "outroot",
            "--noprebin",
            "--nonorm",
            "--radius",
            "3.0",
            "--sigma",
            "1.5",
            "--kernelwidth",
            "2.0",
            "--spherical",
            "--index1",
            "2",
            "--index2",
            "3",
            "--debug",
        ]
    )
    assert args.prebin is False
    assert args.norm is False
    assert args.radius == 3.0
    assert args.sigma == 1.5
    assert args.kernelwidth == 2.0
    assert args.spherical is True
    assert args.index1 == 2
    assert args.index2 == 3
    assert args.debug is True


# ==================== getneighborhood tests ====================


def getneighborhood_cubic_center(debug=False):
    """Test cubic neighborhood extraction at center of volume."""
    if debug:
        print("getneighborhood_cubic_center")
    _reset_getneighborhood_globals()

    xsize, ysize, zsize = 10, 10, 10
    data = np.ones((xsize, ysize, zsize), dtype=np.float64)
    radius = 1.0

    result = getneighborhood(
        data,
        5,
        5,
        5,
        xsize,
        ysize,
        zsize,
        radius,
        spherical=False,
        kernelwidth=None,
    )
    # With radius=1, cubic neighborhood is (2*1+1)^3 = 27 voxels
    assert len(result) == 27
    # All values are 1.0 (kernel is all 1.0 when kernelwidth=None)
    assert np.allclose(result, 1.0)


def getneighborhood_cubic_with_kernel(debug=False):
    """Test cubic neighborhood with gaussian kernel weighting."""
    if debug:
        print("getneighborhood_cubic_with_kernel")
    _reset_getneighborhood_globals()

    xsize, ysize, zsize = 10, 10, 10
    data = np.ones((xsize, ysize, zsize), dtype=np.float64)
    radius = 1.0

    result = getneighborhood(
        data,
        5,
        5,
        5,
        xsize,
        ysize,
        zsize,
        radius,
        spherical=False,
        kernelwidth=1.5,
    )
    # Should have 27 elements
    assert len(result) == 27
    # Since data is all 1.0, the result is the kernel values
    # The max value should be 1.0 (Gaussian peak at distance 0)
    assert np.max(result) == pytest.approx(1.0, abs=0.01)
    # Corner values (distance=sqrt(3)) should be less than center
    assert np.min(result) < np.max(result)


def getneighborhood_cubic_edge(debug=False):
    """Test cubic neighborhood at edge of volume (clipped)."""
    if debug:
        print("getneighborhood_cubic_edge")
    _reset_getneighborhood_globals()

    xsize, ysize, zsize = 10, 10, 10
    data = np.ones((xsize, ysize, zsize), dtype=np.float64)
    radius = 2.0

    # At corner (0,0,0) with radius=2, neighborhood is clipped
    result = getneighborhood(
        data,
        0,
        0,
        0,
        xsize,
        ysize,
        zsize,
        radius,
        spherical=False,
        kernelwidth=None,
    )
    # At corner, we get 0:3 in each dimension = 3^3 = 27 voxels (clipped)
    expected_size = 3 * 3 * 3  # 0 to radius+1 in each dim
    assert len(result) == expected_size


def getneighborhood_cubic_values(debug=False):
    """Test that cubic neighborhood extracts correct values."""
    if debug:
        print("getneighborhood_cubic_values")
    _reset_getneighborhood_globals()

    xsize, ysize, zsize = 10, 10, 10
    data = np.zeros((xsize, ysize, zsize), dtype=np.float64)
    # Set a known pattern
    data[4:7, 4:7, 4:7] = 1.0
    radius = 1.0

    result = getneighborhood(
        data,
        5,
        5,
        5,
        xsize,
        ysize,
        zsize,
        radius,
        spherical=False,
        kernelwidth=None,
    )
    # The entire (2*1+1)^3 = 3^3 block should be 1.0
    assert np.allclose(result, 1.0)


def getneighborhood_spherical_center(debug=False):
    """Test spherical neighborhood extraction at center of volume."""
    if debug:
        print("getneighborhood_spherical_center")
    _reset_getneighborhood_globals()

    xsize, ysize, zsize = 10, 10, 10
    data = np.ones((xsize, ysize, zsize), dtype=np.float64)
    radius = 2.0

    result = getneighborhood(
        data,
        5,
        5,
        5,
        xsize,
        ysize,
        zsize,
        radius,
        spherical=True,
    )
    # Spherical neighborhood with radius=2 should have fewer voxels
    # than cubic (5^3=125), but more than just the center
    assert len(result) > 1
    assert len(result) < 125
    # All values should be 1.0
    assert np.allclose(result, 1.0)


def getneighborhood_spherical_edge(debug=False):
    """Test spherical neighborhood at edge is clipped to valid indices."""
    if debug:
        print("getneighborhood_spherical_edge")
    _reset_getneighborhood_globals()

    xsize, ysize, zsize = 10, 10, 10
    data = np.ones((xsize, ysize, zsize), dtype=np.float64)
    radius = 2.0

    # At edge, should get fewer voxels than at center
    result_center = getneighborhood(
        data,
        5,
        5,
        5,
        xsize,
        ysize,
        zsize,
        radius,
        spherical=True,
    )
    _reset_getneighborhood_globals()
    result_edge = getneighborhood(
        data,
        0,
        0,
        0,
        xsize,
        ysize,
        zsize,
        radius,
        spherical=True,
    )
    assert len(result_edge) < len(result_center)


def getneighborhood_spherical_values(debug=False):
    """Test spherical neighborhood returns correct data values."""
    if debug:
        print("getneighborhood_spherical_values")
    _reset_getneighborhood_globals()

    xsize, ysize, zsize = 10, 10, 10
    rng = np.random.RandomState(42)
    data = rng.randn(xsize, ysize, zsize).astype(np.float64)
    radius = 1.0

    result = getneighborhood(
        data,
        5,
        5,
        5,
        xsize,
        ysize,
        zsize,
        radius,
        spherical=True,
    )
    # With radius=1, spherical neighborhood includes center + 6 face neighbors = 7
    assert len(result) == 7
    # Center value should be included
    assert data[5, 5, 5] in result


def getneighborhood_radius_size_increases(debug=False):
    """Test that larger radius produces larger neighborhood."""
    if debug:
        print("getneighborhood_radius_size_increases")
    xsize, ysize, zsize = 15, 15, 15
    data = np.ones((xsize, ysize, zsize), dtype=np.float64)

    _reset_getneighborhood_globals()
    result_r1 = getneighborhood(
        data,
        7,
        7,
        7,
        xsize,
        ysize,
        zsize,
        1.0,
        spherical=True,
    )
    _reset_getneighborhood_globals()
    result_r2 = getneighborhood(
        data,
        7,
        7,
        7,
        xsize,
        ysize,
        zsize,
        2.0,
        spherical=True,
    )
    _reset_getneighborhood_globals()
    result_r3 = getneighborhood(
        data,
        7,
        7,
        7,
        xsize,
        ysize,
        zsize,
        3.0,
        spherical=True,
    )
    assert len(result_r1) < len(result_r2) < len(result_r3)


# ==================== getMI tests ====================


def getMI_identical_signals(debug=False):
    """Test getMI with identical signals gives high MI."""
    if debug:
        print("getMI_identical_signals")
    _reset_getMI_globals()

    rng = np.random.RandomState(42)
    x = rng.randn(100).astype(np.float64)
    y = x.copy()

    mi = getMI(x, y, norm=True, bins=10, init=True, prebin=True)
    assert mi is not None
    assert mi > 0.0


def getMI_independent_signals(debug=False):
    """Test getMI with independent signals gives lower MI than identical."""
    if debug:
        print("getMI_independent_signals")
    _reset_getMI_globals()

    rng = np.random.RandomState(42)
    x = rng.randn(200).astype(np.float64)
    y = rng.randn(200).astype(np.float64)

    mi_indep = getMI(x, y, norm=True, bins=10, init=True, prebin=True)

    _reset_getMI_globals()
    mi_same = getMI(x, x.copy(), norm=True, bins=10, init=True, prebin=True)

    assert (
        mi_same > mi_indep
    ), f"MI of identical ({mi_same}) should exceed MI of independent ({mi_indep})"


def getMI_correlated_signals(debug=False):
    """Test getMI with correlated signals gives higher MI than independent."""
    if debug:
        print("getMI_correlated_signals")
    _reset_getMI_globals()

    rng = np.random.RandomState(42)
    x = rng.randn(200).astype(np.float64)
    y = 0.8 * x + 0.2 * rng.randn(200).astype(np.float64)

    mi_corr = getMI(x, y, norm=True, bins=10, init=True, prebin=True)

    _reset_getMI_globals()
    y_indep = rng.randn(200).astype(np.float64)
    mi_indep = getMI(x, y_indep, norm=True, bins=10, init=True, prebin=True)

    assert (
        mi_corr > mi_indep
    ), f"MI of correlated ({mi_corr}) should exceed MI of independent ({mi_indep})"


def getMI_no_norm(debug=False):
    """Test getMI without normalization."""
    if debug:
        print("getMI_no_norm")
    _reset_getMI_globals()

    rng = np.random.RandomState(42)
    x = rng.randn(100).astype(np.float64)
    y = x.copy()

    mi = getMI(x, y, norm=False, bins=10, init=True, prebin=True)
    assert mi is not None
    assert mi > 0.0


def getMI_no_prebin(debug=False):
    """Test getMI with prebin=False uses slow path."""
    if debug:
        print("getMI_no_prebin")
    _reset_getMI_globals()

    rng = np.random.RandomState(42)
    x = rng.randn(100).astype(np.float64)
    y = x.copy()

    mi = getMI(x, y, norm=True, bins=10, init=True, prebin=False)
    assert mi is not None
    assert mi > 0.0


def getMI_auto_bins(debug=False):
    """Test getMI with bins=-1 auto-selects bin count."""
    if debug:
        print("getMI_auto_bins")
    _reset_getMI_globals()

    rng = np.random.RandomState(42)
    x = rng.randn(200).astype(np.float64)
    y = x.copy()

    mi = getMI(x, y, norm=True, bins=-1, init=True, prebin=True)
    assert mi is not None
    assert mi > 0.0


def getMI_nonnegative(debug=False):
    """Test getMI always returns non-negative values."""
    if debug:
        print("getMI_nonnegative")
    _reset_getMI_globals()

    rng = np.random.RandomState(42)
    for _ in range(5):
        x = rng.randn(100).astype(np.float64)
        y = rng.randn(100).astype(np.float64)
        mi = getMI(x, y, norm=True, bins=10, init=True, prebin=True)
        assert mi >= 0.0, f"MI should be non-negative, got {mi}"
        _reset_getMI_globals()


# ==================== spatialmi tests ====================


def _run_spatialmi(image1, image2, mask1, mask2, args, image1_4d=False, image2_4d=False):
    """Helper to run spatialmi with mocked IO and CLI parsing. Returns saved files dict."""
    xsize, ysize, numslices = image1.shape[:3]

    if image1_4d:
        timepoints1 = image1.shape[3]
        data1 = image1
    else:
        timepoints1 = 1
        data1 = image1

    if image2_4d:
        timepoints2 = image2.shape[3]
        data2 = image2
    else:
        timepoints2 = 1
        data2 = image2

    dims1 = _make_dims(xsize, ysize, numslices, timepoints1)
    dims2 = _make_dims(xsize, ysize, numslices, timepoints2)
    mask_dims = _make_dims(xsize, ysize, numslices, 1)
    sizes = _make_sizes()

    hdr1 = _make_mock_hdr(xsize, ysize, numslices, timepoints1)
    hdr2 = _make_mock_hdr(xsize, ysize, numslices, timepoints2)
    mask_hdr1 = _make_mock_hdr(xsize, ysize, numslices, 1)
    mask_hdr2 = _make_mock_hdr(xsize, ysize, numslices, 1)

    saved_nifti = {}

    read_count = {"n": 0}

    def mock_readfromnifti(fname, **kwargs):
        read_count["n"] += 1
        if read_count["n"] == 1:
            return (MagicMock(), data1, hdr1, dims1, sizes)
        elif read_count["n"] == 2:
            return (MagicMock(), mask1, mask_hdr1, mask_dims, sizes)
        elif read_count["n"] == 3:
            return (MagicMock(), data2, hdr2, dims2, sizes)
        elif read_count["n"] == 4:
            return (MagicMock(), mask2, mask_hdr2, mask_dims, sizes)

    def mock_savetonifti(arr, hdr_arg, fname, **kwargs):
        saved_nifti[fname] = arr.copy()

    _reset_getneighborhood_globals()
    _reset_getMI_globals()

    with (
        patch(
            "rapidtide.workflows.spatialmi.tide_io.readfromnifti", side_effect=mock_readfromnifti
        ),
        patch("rapidtide.workflows.spatialmi.tide_io.checkspacedimmatch", return_value=True),
        patch("rapidtide.workflows.spatialmi.tide_io.savetonifti", side_effect=mock_savetonifti),
    ):

        spatialmi(args)

    return saved_nifti


def spatialmi_basic_3d(debug=False):
    """Test basic spatialmi workflow with 3D inputs produces output."""
    if debug:
        print("spatialmi_basic_3d")
    xsize, ysize, numslices = 7, 7, 5
    rng = np.random.RandomState(42)
    image1 = rng.randn(xsize, ysize, numslices).astype(np.float64)
    image2 = rng.randn(xsize, ysize, numslices).astype(np.float64)
    mask1 = np.ones((xsize, ysize, numslices), dtype=np.float64)
    mask2 = np.ones((xsize, ysize, numslices), dtype=np.float64)

    args = _make_default_args()
    args.radius = 1.0

    saved = _run_spatialmi(image1, image2, mask1, mask2, args)

    assert "/tmp/test_spatialmi_result" in saved
    assert saved["/tmp/test_spatialmi_result"].shape == (xsize, ysize, numslices)


def spatialmi_identical_images(debug=False):
    """Test that identical images produce higher MI than different images."""
    if debug:
        print("spatialmi_identical_images")
    xsize, ysize, numslices = 7, 7, 5
    rng = np.random.RandomState(42)
    image1 = rng.randn(xsize, ysize, numslices).astype(np.float64) * 10.0
    mask1 = np.ones((xsize, ysize, numslices), dtype=np.float64)
    mask2 = np.ones((xsize, ysize, numslices), dtype=np.float64)

    # Identical images
    args = _make_default_args(outputroot="/tmp/test_spatialmi_same")
    args.radius = 1.0
    saved_same = _run_spatialmi(image1, image1.copy(), mask1, mask2, args)

    # Different images
    image2 = rng.randn(xsize, ysize, numslices).astype(np.float64) * 10.0
    args2 = _make_default_args(outputroot="/tmp/test_spatialmi_diff")
    args2.radius = 1.0
    saved_diff = _run_spatialmi(image1, image2, mask1, mask2, args2)

    mi_same = saved_same["/tmp/test_spatialmi_same_result"]
    mi_diff = saved_diff["/tmp/test_spatialmi_diff_result"]

    # Mean MI should be higher for identical images
    assert np.mean(mi_same) > np.mean(
        mi_diff
    ), f"Mean MI for identical ({np.mean(mi_same)}) should exceed different ({np.mean(mi_diff)})"


def spatialmi_output_nonnegative(debug=False):
    """Test that output MI values are non-negative after nan_to_num."""
    if debug:
        print("spatialmi_output_nonnegative")
    xsize, ysize, numslices = 7, 7, 5
    rng = np.random.RandomState(42)
    image1 = rng.randn(xsize, ysize, numslices).astype(np.float64)
    image2 = rng.randn(xsize, ysize, numslices).astype(np.float64)
    mask1 = np.ones((xsize, ysize, numslices), dtype=np.float64)
    mask2 = np.ones((xsize, ysize, numslices), dtype=np.float64)

    args = _make_default_args()
    args.radius = 1.0

    saved = _run_spatialmi(image1, image2, mask1, mask2, args)

    result = saved["/tmp/test_spatialmi_result"]
    assert np.all(np.isfinite(result)), "Output should have no NaN/Inf values"
    # MI values should be non-negative (allow tiny floating point noise)
    assert np.all(result >= -1e-10), f"MI values should be non-negative, min={np.min(result)}"


def spatialmi_with_mask(debug=False):
    """Test that masked-out voxels have zero MI in output."""
    if debug:
        print("spatialmi_with_mask")
    xsize, ysize, numslices = 7, 7, 5
    rng = np.random.RandomState(42)
    image1 = rng.randn(xsize, ysize, numslices).astype(np.float64) * 10.0
    image2 = rng.randn(xsize, ysize, numslices).astype(np.float64) * 10.0
    mask1 = np.ones((xsize, ysize, numslices), dtype=np.float64)
    mask2 = np.ones((xsize, ysize, numslices), dtype=np.float64)
    # Mask out a region
    mask1[0, :, :] = 0.0

    args = _make_default_args()
    args.radius = 1.0

    saved = _run_spatialmi(image1, image2, mask1, mask2, args)

    result = saved["/tmp/test_spatialmi_result"]
    # Where totalmask (mask1 * mask2) is 0, output should be 0
    assert np.all(result[0, :, :] == 0.0), "Masked-out voxels should have zero MI"


def spatialmi_radius_check(debug=False):
    """Test that radius < 1.0 causes exit."""
    if debug:
        print("spatialmi_radius_check")
    args = _make_default_args()
    args.radius = 0.5

    with pytest.raises(SystemExit):
        spatialmi(args)


def spatialmi_dim_mismatch_input_mask(debug=False):
    """Test that dimension mismatch between input and mask causes exit."""
    if debug:
        print("spatialmi_dim_mismatch_input_mask")
    xsize, ysize, numslices = 7, 7, 5
    image1 = np.ones((xsize, ysize, numslices), dtype=np.float64)

    dims = _make_dims(xsize, ysize, numslices, 1)
    sizes = _make_sizes()
    hdr = _make_mock_hdr(xsize, ysize, numslices, 1)

    def mock_readfromnifti(fname, **kwargs):
        return (MagicMock(), image1, hdr, dims, sizes)

    call_count = {"n": 0}

    def mock_checkspacedimmatch(d1, d2, **kwargs):
        call_count["n"] += 1
        if call_count["n"] == 1:
            return False  # input1 vs mask1 mismatch
        return True

    args = _make_default_args()

    with (
        patch(
            "rapidtide.workflows.spatialmi.tide_io.readfromnifti", side_effect=mock_readfromnifti
        ),
        patch(
            "rapidtide.workflows.spatialmi.tide_io.checkspacedimmatch",
            side_effect=mock_checkspacedimmatch,
        ),
    ):
        with pytest.raises(SystemExit):
            spatialmi(args)


def spatialmi_4d_input(debug=False):
    """Test spatialmi with 4D input selects correct timepoint."""
    if debug:
        print("spatialmi_4d_input")
    xsize, ysize, numslices = 7, 7, 5
    rng = np.random.RandomState(42)
    # 4D data with 3 timepoints
    data_4d = rng.randn(xsize, ysize, numslices, 3).astype(np.float64) * 10.0
    image2 = rng.randn(xsize, ysize, numslices).astype(np.float64) * 10.0
    mask1 = np.ones((xsize, ysize, numslices), dtype=np.float64)
    mask2 = np.ones((xsize, ysize, numslices), dtype=np.float64)

    args = _make_default_args()
    args.radius = 1.0
    args.index1 = 1  # Select second timepoint

    saved = _run_spatialmi(data_4d, image2, mask1, mask2, args, image1_4d=True)

    assert "/tmp/test_spatialmi_result" in saved
    assert saved["/tmp/test_spatialmi_result"].shape == (xsize, ysize, numslices)


def spatialmi_spherical_neighborhood(debug=False):
    """Test spatialmi with spherical neighborhood option."""
    if debug:
        print("spatialmi_spherical_neighborhood")
    xsize, ysize, numslices = 7, 7, 5
    rng = np.random.RandomState(42)
    image1 = rng.randn(xsize, ysize, numslices).astype(np.float64) * 10.0
    image2 = rng.randn(xsize, ysize, numslices).astype(np.float64) * 10.0
    mask1 = np.ones((xsize, ysize, numslices), dtype=np.float64)
    mask2 = np.ones((xsize, ysize, numslices), dtype=np.float64)

    args = _make_default_args()
    args.radius = 1.0
    args.spherical = True

    saved = _run_spatialmi(image1, image2, mask1, mask2, args)

    assert "/tmp/test_spatialmi_result" in saved
    assert saved["/tmp/test_spatialmi_result"].shape == (xsize, ysize, numslices)


# ==================== Main test function ====================


def test_spatialmi(debug=False):
    # _get_parser tests
    if debug:
        print("Running parser tests")
    parser_basic(debug=debug)
    parser_required_args(debug=debug)
    parser_defaults(debug=debug)
    parser_optional_args(debug=debug)

    # getneighborhood tests
    if debug:
        print("Running getneighborhood tests")
    getneighborhood_cubic_center(debug=debug)
    getneighborhood_cubic_with_kernel(debug=debug)
    getneighborhood_cubic_edge(debug=debug)
    getneighborhood_cubic_values(debug=debug)
    getneighborhood_spherical_center(debug=debug)
    getneighborhood_spherical_edge(debug=debug)
    getneighborhood_spherical_values(debug=debug)
    getneighborhood_radius_size_increases(debug=debug)

    # getMI tests
    if debug:
        print("Running getMI tests")
    getMI_identical_signals(debug=debug)
    getMI_independent_signals(debug=debug)
    getMI_correlated_signals(debug=debug)
    getMI_no_norm(debug=debug)
    getMI_no_prebin(debug=debug)
    getMI_auto_bins(debug=debug)
    getMI_nonnegative(debug=debug)

    # spatialmi workflow tests
    if debug:
        print("Running spatialmi workflow tests")
    spatialmi_basic_3d(debug=debug)
    spatialmi_identical_images(debug=debug)
    spatialmi_output_nonnegative(debug=debug)
    spatialmi_with_mask(debug=debug)
    spatialmi_radius_check(debug=debug)
    spatialmi_dim_mismatch_input_mask(debug=debug)
    spatialmi_4d_input(debug=debug)
    spatialmi_spherical_neighborhood(debug=debug)


if __name__ == "__main__":
    test_spatialmi(debug=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import copy
from argparse import Namespace
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray

import rapidtide.filter as tide_filt
import rapidtide.io as tide_io


def _get_parser() -> Any:
    """
    Create and configure argument parser for phase projection to velocity map conversion.

    This function sets up the command line argument parser for the proj2flow tool,
    which converts phase projection movies to velocity maps. It defines required
    positional arguments for input and output files, as well as optional debugging
    flag.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser object with defined command line arguments

    Notes
    -----
    The function currently has commented-out frequency filter arguments (lowestfreq
    and highestfreq) that can be enabled if needed for specific use cases.

    Examples
    --------
    >>> parser = _get_parser()
    >>> args = parser.parse_args(['input.nii', 'output_root'])
    >>> print(args.inputfilename)
    'input.nii'
    >>> print(args.outputroot)
    'output_root'
    """
    # get the command line parameters
    parser = argparse.ArgumentParser(
        prog="proj2flow",
        description="Convert phase projection movie to velocity map.",
        allow_abbrev=False,
    )
    parser.add_argument("inputfilename", type=str, help="The name of the input nifti file.")
    parser.add_argument("outputroot", type=str, help="The root name of the output nifti files.")

    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="Turn on debugging information.",
        default=False,
    )

    """parser.add_argument(
        "lowestfreq",
        type=float,
        help="The low passband frequency limit in Hz (set less than zero to disable HPF).",
    )
    parser.add_argument(
        "highestfreq",
        type=float,
        help="The high passband frequency limit in Hz (set less than zero to disable LPF)",
    )"""

    return parser


def proj2flow(args: Any) -> None:
    """
    Compute 3D velocity fields from projected 4D fMRI data using forward and backward differences.

    This function reads 4D NIfTI fMRI data, computes velocity fields at each timepoint by
    calculating forward and backward differences in spatial and temporal dimensions, and
    saves the resulting velocity fields as 3D NIfTI files. It also saves an average velocity
    field across all timepoints.

    Parameters
    ----------
    args : argparse.Namespace
        Command-line arguments parsed by `_get_parser()`. Expected attributes include:
        - `inputfilename` : str
            Path to the input NIfTI file containing 4D fMRI data.
        - `outputroot` : str
            Root name for output NIfTI files.
        - `debug` : bool, optional
            If True, enables debug printing and additional checks.

    Returns
    -------
    None
        This function does not return a value. It writes NIfTI files to disk.

    Notes
    -----
    - The function assumes the input data is in the same spatial and temporal dimensions
      as specified by the NIfTI header.
    - Forward and backward differences are computed using a 3x3x3 neighborhood in space
      and a single timepoint difference.
    - Velocity fields are saved with intent code 2003 (indicating a vector field).
    - The output files are named using the format:
        - `{outputroot}_{timepoint:02d}.nii.gz` for each timepoint
        - `{outputroot}_average.nii.gz` for the average velocity field

    Examples
    --------
    >>> import argparse
    >>> args = argparse.Namespace(inputfilename='fmri_data.nii.gz',
    ...                           outputroot='velocity',
    ...                           debug=False)
    >>> proj2flow(args)
    """
    # set default variable values
    displayplots = False

    # get the command line parameters
    try:
        args = _get_parser().parse_args()
    except SystemExit:
        _get_parser().print_help()
        raise

    # get the input TR
    inputtr_fromfile, numinputtrs = tide_io.fmritimeinfo(args.inputfilename)
    print("input data: ", numinputtrs, " timepoints, tr = ", inputtr_fromfile)

    input_img, input_data, input_hdr, thedims, thesizes = tide_io.readfromnifti(args.inputfilename)
    if input_hdr.get_xyzt_units()[1] == "msec":
        tr = thesizes[4] / 1000.0
    else:
        tr = thesizes[4]
    Fs = 1.0 / tr
    print("tr from header =", tr, ", sample frequency is ", Fs)

    xsize, ysize, numslices, timepoints = tide_io.parseniftidims(thedims)
    xdim, ydim, slicethickness, tr = tide_io.parseniftisizes(thesizes)
    forwarddiffs = np.zeros((xsize - 2, ysize - 2, numslices - 2, 3, 3, 3), dtype="float32")
    backwarddiffs = np.zeros((xsize - 2, ysize - 2, numslices - 2, 3, 3, 3), dtype="float32")
    velocities = np.zeros((xsize, ysize, numslices, timepoints, 3), dtype="float32")
    averagevelocity = np.zeros((xsize, ysize, numslices, 3), dtype="float32")

    # cycle over all timepoints
    print("now cycling over all timepoints")
    for timepoint in range(timepoints):
        print(f"processing timepoint {timepoint}")
        # calculate forward differences
        for zoffset in [-1, 0, 1]:
            for yoffset in [-1, 0, 1]:
                for xoffset in [-1, 0, 1]:
                    if args.debug:
                        print([xoffset * xdim, yoffset * ydim, zoffset * slicethickness])
                    distance = np.linalg.norm(
                        np.array([xoffset * xdim, yoffset * ydim, zoffset * slicethickness])
                    )
                    # if distance == 0.0:
                    #    distance = 1.0
                    if args.debug:
                        print(f"{xoffset}, {yoffset}, {zoffset}")
                        print(f"\tdistance:{distance}")
                        print(
                            f"\t{input_data[1 + xoffset : xsize + xoffset - 1,1 + yoffset : ysize + yoffset - 1,1 + zoffset : numslices + zoffset - 1,(timepoint + 1) % timepoints,].shape}"
                        )
                        print(
                            f"\t{input_data[1: -1 ,1 : -1,1: -1 ,(timepoint + 1) % timepoints,].shape}"
                        )
                    forwarddiffs[:, :, :, xoffset + 1, yoffset + 1, zoffset + 1] = (
                        (
                            input_data[
                                1 + xoffset : xsize + xoffset - 1,
                                1 + yoffset : ysize + yoffset - 1,
                                1 + zoffset : numslices + zoffset - 1,
                                (timepoint + 1) % timepoints,
                            ]
                            - input_data[
                                1:-1,
                                1:-1,
                                1:-1,
                                timepoint,
                            ]
                        )
                    ) * distance
                    backwarddiffs[:, :, :, xoffset + 1, yoffset + 1, zoffset + 1] = (
                        (
                            input_data[
                                1:-1,
                                1:-1,
                                1:-1,
                                timepoint,
                            ]
                            - input_data[
                                1 + xoffset : xsize + xoffset - 1,
                                1 + yoffset : ysize + yoffset - 1,
                                1 + zoffset : numslices + zoffset - 1,
                                (timepoint - 1) % timepoints,
                            ]
                        )
                    ) * distance
                    if args.debug:
                        print(f"{np.min(forwarddiffs)=}, {np.max(forwarddiffs)=}")
                        print(f"{np.min(backwarddiffs)=}, {np.max(backwarddiffs)=}")
                        print(f"{forwarddiffs.dtype=}")
        velocities[1:-1, 1:-1, 1:-1, timepoint, 0] = np.sum(
            forwarddiffs[:, :, :, :, 1, 1], axis=3
        ) + np.sum(backwarddiffs[:, :, :, :, 1, 1], axis=3)
        velocities[1:-1, 1:-1, 1:-1, timepoint, 1] = np.sum(
            forwarddiffs[:, :, :, 1, :, 1], axis=3
        ) + np.sum(backwarddiffs[:, :, :, 1, :, 1], axis=3)
        velocities[1:-1, 1:-1, 1:-1, timepoint, 2] = np.sum(
            forwarddiffs[:, :, :, 1, 1, :], axis=3
        ) + np.sum(backwarddiffs[:, :, :, 1, 1, :], axis=3)

        # now do the ones with other numbers of time points
        output_hdr = copy.deepcopy(input_hdr)
        output_hdr.set_intent(2003)
        output_hdr["dim"][0] = 4
        output_hdr["dim"][4] = 3
        output_hdr["dim"][5] = 1
        output_hdr.datatype = 16
        if args.debug:
            print(f"{velocities.dtype=}")
        tide_io.savetonifti(
            velocities[:, :, :, timepoint, :].reshape((xsize, ysize, numslices, 3)),
            output_hdr,
            f"{args.outputroot}_{str(timepoint).zfill(2)}",
            debug=args.debug,
        )
        averagevelocity += velocities[:, :, :, timepoint, :].reshape((xsize, ysize, numslices, 3))
    tide_io.savetonifti(
        averagevelocity[:, :, :, :] / timepoints,
        output_hdr,
        f"{args.outputroot}_average",
        debug=args.debug,
    )
